import bs4
from sangfroid.value.value import Value

@Value.handles_type()
class Canvas(Value):
    @property
    def value(self):

        from sangfroid.layer.layer import Layer

        layers = [field
                 for field in self._tag.children
                 if isinstance(field, bs4.element.Tag)
                 ]
        if len([n for n in layers if n.name!='layer'])!=0:
            raise ValueError(
                    f"Only layers can be the children of a canvas: {self._tag}"
                    )

        result = [Layer.from_tag(layer) for layer in layers]

        return result



import os
import sangfroid
import gzip
import pytest
from test import *
from bs4 import BeautifulSoup

GZIP_HEADER = b'\x1f\x8b'

def test_animation_load_sif():
    sif = get_animation('circles.sif')

    assert sif.tag is not None
    assert sif.name == 'Circles'
    assert sif.desc == 'I like circles. They are round.'
    assert sif.size == (480, 270)
    assert sif.xres == 2834.645669
    assert sif.yres == 2835.0
    assert sif.background==sangfroid.value.Color('#808080')
    assert sif.begin_time==0
    assert sif.end_time==sangfroid.T('5s', ref=sif.tag)

def test_animation_load_sifz():
    sif = get_animation('wombats.sifz')

    assert sif.name == 'wombats'
    assert sif.desc == 'I like wombats. They live in Australia.'

def normalise_xml(s):
    # This function used to strip out all space around NavigableStrings
    # and return the changed XML. But that's almost impossible to read
    # in pytest's output. So now it returns a list of:
    #  - for NavigableStrings: the stripped form of item.text
    #           (except that if the stripped form is the empty string,
    #           it's are silently dropped);
    #  - for everything else: lists of name and attributes.

    if s.startswith(GZIP_HEADER):
        s = gzip.decompress(s)

    if isinstance(s, (str, bytes)):
        s = BeautifulSoup(s, features='xml')

    result = []

    for item in s.descendants:
        if isinstance(item, bs4.NavigableString):
            s = item.text.strip()
            if s!='':
                result.append(s)
        elif not item.attrs:
            result.append(item.name)
        else:
            addendum = (
                    item.name + ' ' +
                    ' '.join([
                        f'{k}={v}'
                        for k,v
                        in sorted(item.attrs.items())
                        ]))

            result.append(addendum)

    return result

def test_animation_save_and_saveas():

    with open(os.path.join(
        os.path.dirname(__file__),
        'circles.sif',
        ), 'rb') as f:

        original = f.read()

    with open(os.path.join(
        os.path.dirname(__file__),
        f'purple-circles.sif',
        ), 'rb') as f:

        expected = f.read()

    expected = normalise_xml(expected)

    for compress_source in [False, True]:

        for save_as in [False, True]:

            # Firstly, let's set up the source file that we're
            # going to read back in.

            if compress_source:
                source_suffix = '.sifz'
                source_contents = gzip.compress(original)
            else:
                source_suffix = '.sif'
                source_contents = original

            source_filename = temp_filename(suffix=source_suffix)

            # For assertion messages:
            details = (
                    f'compress_source={compress_source}; '
                    f'save_as={save_as}; '
                    f'source_filename={source_filename}'
                    )

            with open(source_filename, 'wb') as f:
                f.write(source_contents)

            # Right. Now, what happens when we load it, make
            # some changes, and write it out again?

            animation = sangfroid.open(source_filename)

            for circle in animation.find_all('circle'):
                circle['color'].value = '#ff00ff'

            if save_as:
                final_filename = temp_filename()
                details += f'; final_filename={final_filename}'
                animation.save(final_filename)
            else:
                final_filename = source_filename
                animation.save()

            assert os.path.isfile(final_filename), (
                    f"save created a file ({details})"
                    )
            assert os.path.getsize(final_filename)!=0, (
                    f"savefile is not empty ({details})"
                    )

            with open(final_filename, 'rb') as f:
                found = f.read()

            found = normalise_xml(found)

            assert found==expected, (
                    f'{details}; final_filename={final_filename}'
                    )

            os.unlink(source_filename)
            if final_filename!=source_filename:
                os.unlink(final_filename)

def test_animation_saveas_different_format():

    def format_for(suffix):
        if suffix=='sif':
            return sangfroid.format.Sif
        elif suffix=='sifz':
            return sangfroid.format.Sifz
        else:
            raise ValueError(suffix)

    for read_from, write_to in [
            ('sif',  'sif' ),
            ('sif',  'sifz'),
            ('sifz', 'sif' ),
            ('sifz', 'sifz'),
            ]:

        if read_from=='sif':
            source_filename = 'circles.sif'
        elif read_from=='sifz':
            source_filename = 'wombats.sifz'
        else:
            raise ValueError(source_filename)

        source = sangfroid.open(
                os.path.join(
                    os.path.dirname(__file__),
                    source_filename,
                    ))
        assert isinstance(source._format, format_for(read_from))

        tempname = temp_filename(
                suffix=f'.{write_to}')

        source.save(filename=tempname)

        revenant = sangfroid.open(tempname)

        assert source.desc==revenant.desc
        assert isinstance(source._format, format_for(write_to))

        os.unlink(tempname)

def test_animation_len():
    sif = get_animation('bouncing.sif')
    assert len(sif)==121

def test_animation_blank_simple():

    sif = sangfroid.Animation()
    blank_sif_assertions(sif, 'simple')

def blank_sif_assertions(sif, name):

    assert sif.name == 'New animation', name
    assert sif.desc == '', name
    assert sif.size == (480, 270), name
    assert sif.xres == 2834.645669
    assert sif.yres == 2834.645669
    assert sif.background==sangfroid.value.Color('#808080'), name
    assert sif.begin_time==0, name
    assert sif.end_time==sangfroid.T('5s', ref=sif.tag), name

    keyframes = list(sif.keyframes)
    assert len(keyframes)==1, name
    assert keyframes[0].time==0, name

def test_animation_blank_save():
    sif = sangfroid.Animation()
    
    names = {
            'uncompressed': temp_filename(suffix='.sif'),
            'compressed': temp_filename(suffix='.sifz'),
            }

    with pytest.raises(ValueError):
        sif.save()

    for n in names.values():
        sif.save(n)

    for name, filename in names.items():
        sif2 = sangfroid.Animation(filename)
        blank_sif_assertions(sif2, name)



import re
import sangfroid
import pytest
from test import *

def test_layer_children():
    sif = get_animation('pick-and-mix.sif')

    found = [re.search(r'([a-z]+)', str(layer))[0]
             for layer in sif.descendants]

    assert found==[
            "zoom",
            "translate",
            "rotate",
            "timeloop",
            "stroboscope",
            "freetime",
            "shade",
            "bevel",
            "xor",
            "text",
            "switch",
            "super",
            "sound",
            "skeleton",
            "plant",
            "switch",
            "group",
            "filter",
            "duplicate",
            "spiral",
            "radial",
            "noise",
            "linear",
            "curve",
            "conical",
            "star",
            "solid",
            "region",
            "rectangle",
            "polygon",
            "outline",
            "outline",
            "circle",
            "checker",
            "advanced",
            "mandelbrot",
            "julia",
            "lumakey",
            "halftone",
            "halftone",
            "colorcorrect",
            "clamp",
            "chromakey",
            "simple",
            "metaballs",
            "warp",
            "twirl",
            "stretch",
            "spherize",
            "skeleton",
            "noise",
            "inside",
            "curve",
            "radial",
            "motion",
            "blur",
            ]

def test_layer_items(dump = False):
    sif = get_animation('bouncing.sif')

    def format_value(v):
        result = f'{v.__class__.__name__}, {v}'
        if v.is_animated:
            result += ', animated'
        return result

    found = ''
    for layer in sif.descendants:
        found += f'{layer}\n'

        for k, v in layer.items():

            if isinstance(v, sangfroid.value.Composite):
                found += f' - {k}:\n'
                for k2, v2 in v.items():
                    found += f'     - {k2} = '+format_value(v2)+'\n'
            else:
                found += f' - {k}: '+format_value(v)+'\n'

    if dump:
        print(found)
    else:
        assert found == LAYER_ITEMS_EXPECTED

def _find_type_names_of_children_of_layer(layer):
    """
    Given a layer, returns a list of strings, each being the name of a
    layer which is a child of the parameter, in order; if a layer has
    no name, the member is None. The result is in the same order that
    the layers appear in the sif file.

    This gives a sort of signature for checking that the layers we find
    are the layers we expected to find.
    """
    result = [
            [
            b.type_
            for b in a
            ] for a in layers
            ]
    return result

def _find_the_shadows(t):
    return 'shadow' in t.get('desc', '').lower()

def test_layer_find_all_simple():
    sif = get_animation('bouncing.sif')
    shadows = sif.find_all(desc='Shadow')
    assert [x.desc for x in shadows] == ['Shadow', 'Shadow']

    circles = sif.find_all('circle')
    assert [x.desc for x in circles] == [
            'Shadow circle', 'Bouncy ball']

    wombats = list(sif.find_all('wombat'))
    assert wombats == []

    everything = list(sif.find_all(True))
    assert len(everything)==11

    shadows = [x.desc for x in sif.find_all(_find_the_shadows)]
    assert sorted(shadows)==['Shadow', 'Shadow', 'Shadow circle']

    bananas = sif.find_all(desc='Banana')
    assert len(bananas)==0

def test_layer_find_all_recursive():
    sif = get_animation('bouncing.sif')

    shadows = [x.desc for x in sif.find_all(_find_the_shadows,
                                            recursive=False,
                                            )]
    assert sorted(shadows)==['Shadow']

    shadows = [x.desc for x in sif.find_all(_find_the_shadows,
                                            recursive=True,
                                            )]
    assert sorted(shadows)==['Shadow', 'Shadow', 'Shadow circle']

def test_layer_find():
    sif = get_animation('circles.sif')
    orange_circle = sif.find(desc='Orange circle')
    assert isinstance(orange_circle, sangfroid.layer.Layer)
    assert orange_circle.desc == 'Orange circle'

    pink_circle = sif.find(desc='Pink circle')
    assert pink_circle is None, 'Matching nothing gives None'

def test_layer_item_get():
    sif = get_animation('circles.sif')
    green_circle = sif.find(desc='Green circle')
    logger.debug("dir(%s) == %s", green_circle, dir(green_circle))
    assert green_circle.radius == 0.5055338531

    try:
        green_circle.wombat
        ok = False
    except AttributeError:
        ok = True

    assert ok, 'use of unreal attribute should raise KeyError'

def test_layer_item_set():
    sif = get_animation('circles.sif')
    green_circle = sif.find(desc='Green circle')
    assert green_circle.color == '#10FF00', (
            'we can read the colour'
            )

    red = sangfroid.value.Color('#770000')
    green_circle.color = red
    assert green_circle.color == '#770000', (
            'colour has been set to red via Color'
            )

    green_circle.color = '#FFFF00'
    assert green_circle.color == '#FFFF00', (
            'colour has been set to yellow via string'
            )

    with pytest.raises(ValueError):
        green_circle.color = '2s'

    with pytest.raises(TypeError):
        green_circle.color = sangfroid.T(s=2)

def test_layer_item_contains():
    sif = get_animation('circles.sif')
    green_circle = sif.find(desc='Green circle')

    assert 'color' in green_circle
    assert 'wombat' not in green_circle

def test_layer_find_type():

    def names_of(n):
        return set([x.desc for x in n])

    sif = get_animation('circles.sif')

    circles = sif.find_all(type='circle')
    assert names_of(circles) == CIRCLES

    circles = sif.find_all(type=sangfroid.layer.Circle)
    assert names_of(circles) == CIRCLES

    circles = sif.find_all(type='CIRCLE')
    assert names_of(circles) == CIRCLES

    circles = sif.find_all(type='c_i_rcle')
    assert names_of(circles) == CIRCLES

def test_text_simple():
    sif = get_animation('pick-and-mix.sif')

    text = sif.find('text')
    assert text.text=='Hello wombat!'
    assert isinstance(text.text, sangfroid.value.String)

    text.text='Bananas'

    assert text.text=='Bananas'
    assert isinstance(text.text, sangfroid.value.String)

def test_layer_active():
    sif = get_animation('circles.sif')

    black = sif.find(desc='Black circle')
    assert black.active
    assert black.tag['active']=='true'

    black.active = False
    assert not black.active
    assert black.tag['active']=='false'

    black.active = True
    assert black.active
    assert black.tag['active']=='true'

    black.active = True
    assert black.active
    assert black.tag['active']=='true'

LAYER_ITEMS_EXPECTED = """
[🕰️timeloop]
 - z_depth: Real, 0.0
 - link_time: Time, 0f
 - local_time: Time, 0f
 - duration: Time, 2s
 - only_for_positive_duration: Bool, False
 - symmetrical: Bool, True
[📂group 'Ball']
 - z_depth: Real, 0.0
 - amount: Real, 0.75
 - blend_method: Integer, 13
 - origin: X_Y, (0.0, 0.0)
 - transformation:
     - offset = X_Y, (animated), animated
     - angle = Angle, 0°
     - skew_angle = Angle, 0°
     - scale = X_Y, (animated), animated
 - canvas: Canvas, [[-🔵circle 'Bouncy ball'], [-🫴bevel]]
 - color: Color, #ff0000
 - radius: Real, 1.0
 - feather: Real, 0.0
 - invert: Bool, False
 - type: Integer, 1
 - color1: Color, #ffffff
 - color2: Color, #000000
 - angle: Angle, 89.0588°
 - depth: Real, 0.5819661441
 - softness: Real, 0.3276240462
 - use_luma: Bool, False
 - solid: Bool, False
 - fake_origin: X_Y, (0.0, 0.0)
 - time_dilation: Real, 1.0
 - time_offset: Time, 0f
 - children_lock: Bool, False
 - outline_grow: Real, 0.0
 - z_range: Bool, False
 - z_range_position: Real, 0.0
 - z_range_depth: Real, 0.0
 - z_range_blur: Real, 0.0
[-🫴bevel]
 - z_depth: Real, 0.0
 - amount: Real, 0.75
 - blend_method: Integer, 13
 - type: Integer, 1
 - color1: Color, #ffffff
 - color2: Color, #000000
 - angle: Angle, 89.0588°
 - depth: Real, 0.5819661441
 - softness: Real, 0.3276240462
 - use_luma: Bool, False
 - solid: Bool, False
 - fake_origin: X_Y, (0.0, 0.0)
[-🔵circle 'Bouncy ball']
 - z_depth: Real, 0.0
 - amount: Real, 1.0
 - blend_method: Integer, 0
 - color: Color, #ff0000
 - radius: Real, 1.0
 - feather: Real, 0.0
 - origin: X_Y, (0.0, 0.0)
 - invert: Bool, False
[📂group 'Shadow']
 - z_depth: Real, 0.0
 - amount: Real, 1.0
 - blend_method: Integer, 1
 - origin: X_Y, (0.0, -1.6666666269)
 - transformation:
     - offset = X_Y, (0.0, -1.6003249884)
     - angle = Angle, 0°
     - skew_angle = Angle, 0°
     - scale = X_Y, (1.0, 0.3899480104)
 - canvas: Canvas, [[--🔵circle 'Shadow circle'], [--🟠blur]]
 - color: Color, #00000072
 - radius: Real, 1.0
 - feather: Real, 0.0
 - invert: Bool, False
 - size: X_Y, (0.25, 0.25)
 - type: Integer, 1
 - time_dilation: Real, 1.0
 - time_offset: Time, 0f
 - children_lock: Bool, False
 - outline_grow: Real, 0.0
 - z_range: Bool, False
 - z_range_position: Real, 0.0
 - z_range_depth: Real, 0.0
 - z_range_blur: Real, 0.0
[-📂group 'Shadow']
 - z_depth: Real, 0.0
 - amount: Real, 1.0
 - blend_method: Integer, 1
 - origin: X_Y, (0.0, -1.6666666269)
 - transformation:
     - offset = X_Y, (0.0, -1.6003249884)
     - angle = Angle, 0°
     - skew_angle = Angle, 0°
     - scale = X_Y, (1.0, 0.3899480104)
 - canvas: Canvas, [[--🔵circle 'Shadow circle'], [--🟠blur]]
 - color: Color, #00000072
 - radius: Real, 1.0
 - feather: Real, 0.0
 - invert: Bool, False
 - size: X_Y, (0.25, 0.25)
 - type: Integer, 1
 - time_dilation: Real, 1.0
 - time_offset: Time, 0f
 - children_lock: Bool, False
 - outline_grow: Real, 0.0
 - z_range: Bool, False
 - z_range_position: Real, 0.0
 - z_range_depth: Real, 0.0
 - z_range_blur: Real, 0.0
[📂group 'Background']
 - z_depth: Real, 0.0
 - amount: Real, 1.0
 - blend_method: Integer, 0
 - origin: X_Y, (0.0, 0.0)
 - transformation:
     - offset = X_Y, (0.0, 0.0)
     - angle = Angle, 0°
     - skew_angle = Angle, 0°
     - scale = X_Y, (1.0, 1.0)
 - canvas: Canvas, [[-▊solid_color 'wall'], [-🟦rectangle 'floor']]
 - color: Color, #ffb356
 - point1: X_Y, (-4.0130858421, -2.3096354008)
 - point2: X_Y, (4.0234375, -0.9031249881)
 - expand: Real, 0.0
 - invert: Bool, False
 - feather_x: Real, 0.0
 - feather_y: Real, 0.0
 - bevel: Real, 0.0
 - bevCircle: Bool, True
 - time_dilation: Real, 1.0
 - time_offset: Time, 0f
 - children_lock: Bool, False
 - outline_grow: Real, 0.0
 - z_range: Bool, False
 - z_range_position: Real, 0.0
 - z_range_depth: Real, 0.0
 - z_range_blur: Real, 0.0
[-🟦rectangle 'floor']
 - z_depth: Real, 0.0
 - amount: Real, 1.0
 - blend_method: Integer, 0
 - color: Color, #ffb356
 - point1: X_Y, (-4.0130858421, -2.3096354008)
 - point2: X_Y, (4.0234375, -0.9031249881)
 - expand: Real, 0.0
 - invert: Bool, False
 - feather_x: Real, 0.0
 - feather_y: Real, 0.0
 - bevel: Real, 0.0
 - bevCircle: Bool, True
[-▊solid_color 'wall']
 - z_depth: Real, 0.0
 - amount: Real, 1.0
 - blend_method: Integer, 0
 - color: Color, #ffffff
""".lstrip()

CIRCLES = {
        'Black circle', 'Maybe white circle', 'Blue circle',
        'Purple circle', 'Orange circle', 'Green circle',
        'Yellow circle', 'Red circle',
        }

if __name__=='__main__':
    test_layer_items(dump = True)



import bs4
from sangfroid.value.value import Value
from sangfroid.value.color import Color

@Value.handles_type()
class Gradient(Value):
    @property
    def value(self):
        colours = self._tag.find_all('color')
        if len(colours)<2:
            self._raise_colour_count_error()

        result = dict([
                (float(c['pos']), Color(c)) for c in colours
                ])

        return result

    @value.setter
    def value(self, v):
        if isinstance(v, Gradient):
            v = v.value

        if not isinstance(v, dict):
            raise TypeError("Gradient.value must be a dict")

        if len(v)<2:
            self._raise_colour_count_error()

        self._tag.clear()

        for pos, colour in sorted(v.items()):

            colour_tag = Color(colour).tag
            colour_tag['pos'] = '%.06g' % (pos,)
            self._tag.append(colour_tag)

    def _raise_colour_count_error(self):
        raise ValueError("there should be at least two colours in a gradient")

    def __len__(self):
        return len(self.value)

    def __getitem__(self, n):
        return self.value[n]

    def __setitem__(self, n, v):
        previous = self.value
        previous[n] = v
        self.value = previous

    def keys(self):
        return self.value.keys()

    def values(self):
        return self.value.values()

    def items(self):
        return self.value.items()

    def __iter__(self):
        yield from iter(self.value)

    def __str__(self):
        return (
                '{' +
                ','.join([f'{k}:{v}' for k,v in self.items()]) +
                '}')

    def as_python_expression(self):
        return (
                '{' +
                ','.join([f'{k}:{v.as_python_expression()}'
                          for k,v in self.items()]) +
                '}')

        return str(self)



from sangfroid.layer import Layer
import sangfroid.value as v
import sangfroid.field as f

@Layer.handles_type()
class Blur(Layer):
    SYMBOL = '🟠'

    ### {{{
    SYNFIG_VERSION = "0.3"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                        )
    amount               = f.ParamTagField(v.Real, 1.0,
                        )
    blend_method         = f.ParamTagField(v.BlendMethod, v.BlendMethod.STRAIGHT,
                        )
    size                 = f.ParamTagField(v.X_Y, (0.1000000015, 0.1000000015),
                        )
    type_                = f.ParamTagField(v.Integer, 1,
                        )

    ### }}}

@Layer.handles_type()
class Radial_Blur(Layer):
    SYMBOL = '🟠'

    ### {{{
    SYNFIG_VERSION = "0.1"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                        )
    amount               = f.ParamTagField(v.Real, 1.0,
                        )
    blend_method         = f.ParamTagField(v.BlendMethod, v.BlendMethod.STRAIGHT,
                        )
    origin               = f.ParamTagField(v.X_Y, (0.0, 0.0),
                        )
    size                 = f.ParamTagField(v.Real, 0.2,
                        )
    fade_out             = f.ParamTagField(v.Bool, False,
                        )

    ### }}}

@Layer.handles_type()
class Motion_Blur(Layer):
    SYMBOL = '🟠'

    ### {{{
    SYNFIG_VERSION = "0.2"

    aperture             = f.ParamTagField(v.Time, '1s',
                        )
    subsamples_factor    = f.ParamTagField(v.Real, 1.0,
                        )
    subsampling_type     = f.ParamTagField(v.Integer, 2,
                        )
    subsample_start      = f.ParamTagField(v.Real, 0.0,
                        )
    subsample_end        = f.ParamTagField(v.Real, 1.0,
                        )

    ### }}}



import argparse
import pysrt
import os
import sys
import logging
import sangfroid
from bywords.adjust import adjust

logging.basicConfig(stream=sys.stdout)
logger = logging.getLogger('bywords')

def main():
    parser = argparse.ArgumentParser(
            description='make words appear in time with subtitles')
    parser.add_argument(
            'sif', type=str,
            help='name of the .sif animation file')
    parser.add_argument(
            'srt', type=str, nargs='?',
            help='name of the .srt subtitles file')
    parser.add_argument(
            '--letter', '-l', action='store_true',
            help="letter at a time (rather than word at a time)")
    parser.add_argument(
            '--verbose', '-v', action='count', default=0,
            help="print what's being done")
    args = parser.parse_args()

    if args.srt is None:
        srt_filename = os.path.splitext(args.sif)[0]+'.srt'
    else:
        srt_filename = args.srt

    if args.verbose>1:
        logger.setLevel(logging.DEBUG)
    elif args.verbose==1:
        logger.setLevel(logging.INFO)
    else:
        logger.setLevel(logging.WARNING)

    srt = pysrt.open(srt_filename)
    animation = sangfroid.Animation(args.sif)
    logger.debug("Loaded animation and timings.")

    adjust(srt, animation,
           letter_at_a_time = args.letter,
           )

    logger.debug("Saving...")
    animation.save(args.sif)
    logger.info("Done.")

if __name__=='__main__':
    main()



from sangfroid.util import normalise_synfig_layer_type_name

class Registry:
    """
    Decorator, mapping names of things onto classes that handle those things.

    Names are case-insensitive, and underscores are not significant.

    Args to the decorator:
        name (str, optional): the name to associate this class with.
            If None or omitted, the name of the class is used.
    """
    def __init__(self):
        self.handlers = {}

    # Decorator.
    def __call__(self, name=None):
        def _inner(c):
            n = name or c.__name__
            n = normalise_synfig_layer_type_name(n)

            self.handlers[n] = c
            return c
        return _inner

    def from_name(self, name):
        """
        Returns whatever is associated with "name".

        Returns:
            any type

        Raises:
            KeyError, if the name doesn't refer to anything.
        """
        name = normalise_synfig_layer_type_name(name)

        if name not in self.handlers:
            raise KeyError(
                    f"This tag is a {name}, which I don't know how "
                    "to handle.\n\n"
                    "Here are the things I do know:\n"
                    f"  {' '.join(sorted(self.handlers.keys()))}"
                    )
        result = self.handlers[name]

        return result



from sangfroid.value.value import Value
from sangfroid.t import T
import warnings

NEAR_AS_DAMMIT = 0.0001

class Simple(Value):

    our_type = None

    @property
    def value(self):
        if self.our_type is None:
            raise NotImplementedError()

        result = self._tag.get('value', None)
        if result is None:
            raise ValueError(
                    f"This tag should have had a 'value' attribute, "
                    f"but it didn't:\n\n"
                    f"{self._tag}")
        assert isinstance(result, str), f"{result} {type(result)} {self._tag} {type(self._tag)}"

        result = self._construct_value(result)

        return result

    def _construct_value(self, v):
        return self.our_type(v)

    @value.setter
    def value(self, v):
        if self.our_type is None:
            raise NotImplementedError()

        if v==() or v is None:
            result = self.our_type()

        else:
            try:
                result = self.our_type(v)
            except TypeError:
                raise TypeError("I need a value of type "
                                f"{self.our_type.__name__}, "
                                "not "
                                f"{v.__class__.__name__}."
                                )

        result = self._value_to_str(result)

        self._tag.name = self.__class__.__name__.lower()
        self._tag.attrs = {
                'value': result,
                }
        self._tag.clear()

    @classmethod
    def _value_to_str(cls, v):
        return str(v)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            v = other.value
        else:
            try:
                v = self.our_type(other)
            except ValueError:
                v = other

        return self._compare_with(v)

    def _compare_with(self, v):
        return self.value==v

class Numeric(Simple):
    def __float__(self):
        return float(self.value)

    def __int__(self):
        return int(self.value)

@Value.handles_type()
class Real(Numeric):
    our_type = float

    def _compare_with(self, v):
        try:
            return abs(self.value-v)<=NEAR_AS_DAMMIT
        except TypeError:
            return False

@Value.handles_type()
class Integer(Numeric):
    our_type = int

    def __int__(self):
        return self.value

@Value.handles_type()
class Bool(Simple):
    our_type = bool

    def __bool__(self):
        return self.value

    def _construct_value(self, v):
        assert isinstance(v, str)
        if v.lower()=='true':
            return True
        elif v.lower()=='false':
            return False
        else:
            warnings.warn(
                    "boolean string should have been 'true' or 'false', "
                    f"but it was {repr(v)}; treating as False.")
            return False

    @classmethod
    def _value_to_str(cls, v):
        return str(v).lower()

@Value.handles_type()
class Angle(Simple):
    our_type = float

    def _str_inner(self):
        return '%g°' % (self.value,)

    def as_python_expression(self):
        v = self.value
        return str(v)

    def __float__(self):
        return self.value

@Value.handles_type()
class Time(Simple):
    our_type = T

    # We don't provide __float__ and __int__ here because
    # they'd be ambiguous between frames and seconds.

    def _construct_value(self, v):
        return self.our_type(
                v,
                ref = self._tag,
                )

    def as_python_expression(self):
        v = self.value

        if v==0:
            return 0
        else:
            return repr(str(v))



from sangfroid.t import T

class Keyframe:
    def __init__(self, tag):
        self._tag = tag

    @property
    def time(self):
        return T(self._tag['time'],
                 ref = self._tag,
                 )

    @property
    def name(self):
        return self._tag.text

    @property
    def active(self):
        if self._tag['active']=='true':
            return True
        else:
            return False

    def __str__(self):
        if self.active:
            active = ''
        else:
            active = ' (inactive)'
        return f'{self.time} {repr(self.name)}{active}'

    __repr__ = __str__

    @classmethod
    def all_in_animation(cls, animation):
        result = Keyframes.__new__(Keyframes)
        result.animation = animation

        return result

class Keyframes:
    def __init__(self):
        raise ValueError("Don't instantiate the Keyframes class directly.")

    def __len__(self):
        return len(self.animation.tag.find_all('keyframe'))

    def __iter__(self):
        for keyframe_tag in self.animation.tag.find_all('keyframe'):
            yield Keyframe(keyframe_tag)

    def __str__(self):
        result = f'Keyframes of {self.animation}:'
        if len(self)==0:
            result += '\n  (none)'
        else:
            for keyframe in self:
                result += f'\n  - {keyframe}'

        return result

    __repr__ = __str__

__all__ = [
        'Keyframe',
        ]



import copy
import logging
import bs4
import sangfroid.value as v
from sangfroid.registry import Registry
from sangfroid.field import (
        Field,
        TagAttrField,
        ParamTagField,
        TagField,
        )
from sangfroid.util import (
        normalise_synfig_layer_type_name,
        type_and_str_to_value,
        type_and_value_to_str,
        )

logger = logging.getLogger('sangfroid')

class Layer:

    SYMBOL = '?' # fallback

    type_            = TagAttrField(
            str,         None,
            name='type',
            doc = """The name Synfig uses internally for this type of layer.

            In Python, you must spell this as `type_`, because `type` is
            a reserved word.""",
            )
    active           = TagAttrField(
            bool,        True,
            doc = "True if this layer is enabled.",
            )
    exclude_from_rendering = TagAttrField(
            bool,  False,
            doc = "True if this layer should not be rendered.",
            )
    version          = TagAttrField(float,       None)

    desc             = TagAttrField(
            str,         '',
            doc = "A description of this layer.",
            )

    tag              = TagField()

    ########################

    def __init__(self, tag):
        self._tag = tag

    @property
    def parent(self):
        cursor = self._tag.parent
        while cursor is not None:
            if cursor.name=='layer':
                return Layer.from_tag(cursor)
                return cursor
            cursor = cursor.parent

    def __repr__(self):
        result = '['
        result += ('-'*self.tag_depth)
        result += self.SYMBOL
        result += self.__class__.__name__.lower()
        try:
            desc = self.desc

            if desc:
                result += ' ' + repr(desc)
        except KeyError:
            pass

        result += ']'
        return result

    def __getitem__(self, f):
        found = self._tag.find('param', attrs={'name': f})
        if found is None:
            raise KeyError(f)
        return _name_and_value_of(found)[1]

    def __setitem__(self, f, val):
        found = self._tag.find('param', attrs={'name': f})
        if found is None:
            raise KeyError(f)
        old_value = _name_and_value_of(found)[1]

        if isinstance(val, v.Value):
            if not isinstance(val, old_value.__class__):
                raise TypeError(val.__class__)

            new_value = val
        else:
            new_value = old_value.__class__(val)

        old_value.tag.replace_with(new_value.tag)

    def __contains__(self, f):
        found = self._tag.find(
                'param',
                attrs={'name': f},
                )
        return found is not None
    
    @property
    def tag_depth(self):
        cursor = self._tag.parent
        result = 0
        while cursor is not None:
            if cursor.name=='layer':
                result += 1
            cursor = cursor.parent
        return result

    def find_all(self,
                 *args,
                 recursive=True,
                 **kwargs,
                 ):

        matching_special = None

        if len(args)>1:
            raise v.ValueError(
                    "You can only give one positional argument.")
        elif len(args)==1:

            if (
                    isinstance(args[0], str) or
                    (isinstance(args[0], type) and
                     issubclass(args[0], Layer))
                    ):
                if 'type' in kwargs:
                    raise v.ValueError(
                            "You can't give a type in both the positional "
                            "and keyword arguments.")

                kwargs['type'] = args[0]

            elif isinstance(args[0], bool):
                matching_special = args[0]

            elif hasattr(args[0], '__call__'):
                matching_special = args[0]

            else:
                raise TypeError(args[0])

        if 'attrs' in kwargs:
            for k,v in kwargs['attrs'].items():
                if k in kwargs:
                    raise v.ValueError("{k} specified both as a kwarg and in attrs")
                kwargs[k] = v

            del kwargs['attrs']

        for k,v in kwargs.items():
            if k=='type':
                if not isinstance(v, str):
                    v = v.__name__

                kwargs[k] = v.lower().replace('_', '')

        logger.debug("begin find_all")

        def matcher(found_tag):
            if found_tag.name!='layer':
                return False

            logger.debug("considering tag: %s %s",
                         found_tag.name, found_tag.attrs)

            found_layer = Layer.from_tag(found_tag)

            if matching_special is None:

                def munge(k,v):
                    if k=='type':
                        k = 'type_'
                        v = v.lower()

                    return (k,v)

                targets = [
                    munge(k,v)
                    for k,v in kwargs.items()
                    ]

                logger.debug("want: %s", targets)

                for k, want_value in targets:
                    try:
                        found_value = getattr(found_layer, k)
                        logger.debug("  -- %s field is %s; want %s", k,
                                     repr(found_value),
                                     repr(want_value),
                                     )
                    except AttributeError:
                        logger.debug("  -- it does not have a %s", k)
                        continue

                    logger.debug("  want: %s  found: %s",
                                 want_value, found_value)

                    if found_value==want_value:
                        logger.debug("    -- a match!")
                        return True

                logger.debug("  -- no matches.")
                return False

            elif isinstance(matching_special, bool):
                return matching_special

            else:
                result = matching_special(found_tag)
                logger.debug("  -- callback says: %s", result)
                return result

            raise v.ValueError(found_tag)

        result = [
                self.from_tag(x) for x in
                self._tag.find_all(matcher,
                                  recursive=recursive,
                                  )
                ]
        logger.debug("find_all found: %s",
                     result,
                     )

        return result

    @property
    def children(self):
        return
        yield

    def find(self, *args, **kwargs):
        items = self.find_all(*args, **kwargs)
        if items:
            return items[0]
        else:
            return None

    __call__ = find

    ########################

    handles_type = Registry()

    @classmethod
    def from_tag(cls, tag):
        tag_type = tag.get('type', None)
        if tag_type is None:
            raise v.ValueError(
                    f"tag has no 'type' field: {tag}")
        return cls.handles_type.from_name(name=tag_type)(tag)

    def _as_dict(self):
        return dict([
            _name_and_value_of(param)
            for param in self._tag.find_all('param')
            ])

    def items(self):
        return self._as_dict().items()

    def keys(self):
        return self._as_dict().keys()

    def values(self):
        return self._as_dict().values()

    def __iter__(self):
        return self._as_dict().__iter__()

    def _get_param(self, k):
        tag = self._tag.find('param',
                            attribs={
                                'name': k,
                                })
        if tag is None:
            return None

        raise ValueError(f"{k}, {tag}")

def _name_and_value_of(tag):
    if tag.name!='param':
        raise v.ValueError(f"param is not a <param>: {tag}")

    name = tag.get('name', None)
    if name is None:
        raise v.ValueError(f"param has no 'name' field: {tag}")

    value_tags = [tag for tag in tag.children
                  if isinstance(tag, bs4.element.Tag)
                  ]

    if len(value_tags)!=1:
        raise v.ValueError(f"param should have one value: {tag}")

    value_tag = value_tags[0]

    value = v.Value.from_tag(value_tag)
    return name, value

__all__ = [
        'Layer',
        ]



import pytest
import sangfroid
from sangfroid import T
from test import *

M24 = 24*60
H24 = 24*60*60

TESTS = [
        # "None" in secs means not to look it up.

        # Init param; fps;  frames; secs;  str or exception; why
        (None,        None,      0,    0,  ValueError, 'silly time spec'),
        ('',          None,      0,    0,  ValueError, 'silly time spec'),
        ('Banana',    None,      0,    0,  ValueError, 'silly time spec'),
        (2+4j,        None,      0,    0,  ValueError, 'silly time spec'),

        ('0s',        None,      0,    0,  ValueError, 'seconds, no FPS'),
        ('2s',        None,      0,    0,  ValueError, 'seconds, no FPS'),

        ('0s',          -1,      0,    0,  ValueError, 'silly FPS specs'),
        ('0s',        24.1,      0,    0,  ValueError, 'silly FPS specs'),
        ('0s',    'wombat',      0,    0,  ValueError, 'silly FPS specs'),

        ('0s',          24,      0,    0,  '0f',       'seconds'),
        ('2s',          24,     48,    2,  '2s',       'seconds'),
        ('4s',          24,     96,    4,  '4s',       'seconds'),
        ('-2s',         24,    -48,   -2,  '-2s',      'seconds'),
        ('2.5s',        24,     60,  2.5,  '2s 12f',   'seconds'),
        ('-2.5s',       24,    -60, -2.5,  '-2s 12f',  'seconds'),

        ('2m',          24,   2880,  120,  '2m',       'minutes'),
        ('2m 12f',      24,   2892,120.5,  '2m 12f',   'min+f'),
        ('2m 5s',       24,   3000,  125,  '2m 5s',    'min+sec'),
        ('2m 5s 12f',   24,   3012,125.5,  '2m 5s 12f','min+sec+f'),

        ('2h',          24, 172800, 7200,  '2h',       'hours'),
        ('2h 2m',       24, 175680, 7320,  '2h 2m',    'h, m'),
        ('2h 2m 2s',    24, 175728, 7322,  '2h 2m 2s', 'h, m, s'),
        ('2h 2m 2s 2f', 24, 175730,7322.08,'2h 2m 2s 2f','h, m, s, f'),

        ('0s',        None,      0,    0, '0f',        's+f, no fps'),
        ('0s 2f',     None,      2, 0.08,  ValueError, 's+f, no fps'),
        ('2s 2f',     None,      0,    0,  ValueError, 's+f, no fps'),

        ('0s 0f',       24,      0,    0,  '0f',       's+f'),
        ('0s 2f',       24,      2, 0.08,  '2f',       's+f'),
        ('0s 2.5f',     24,    2.5,  0.1,  '2.5f',     's+f'),
        ('0s 100f',     24,    100, 4.17,  '4s 4f',    's+f'),
        ('2s 2f',       24,     50, 2.08,  '2s 2f',    's+f'),
        ('2s -2f',      24,      0,    0,  ValueError, 's+f'),
        ('4s 0f',       24,     96,    4,  '4s',       's+f'),
        ('-2s -2f',     24,      0,    0,  ValueError, 's+f'),
        ('-2s 2f',      24,    -50,-2.08,  '-2s 2f',   's+f'),
        ('2.5s 0f',     24,     60,  2.5,  '2s 12f',   's+f'),
        ('2.5s 2f',     24,     62, 2.58,  '2s 14f',   's+f'),
        ('2.5s 2.5f',   24,   62.5,  2.6,  '2s 14.5f', 's+f'),

        ('0f',        None,      0, None,  '0f',       'f, no fps'),
        ('2f',        None,      2, None,  '2f',       'f, no fps'),
        ('24f',       None,     24, None,  '24f',      'f, no fps'),
        ('48f',       None,     48, None,  '48f',      'f, no fps'),
        ('48.5f',     None,   48.5, None,  '48.5f',    'f, no fps'),
        ('-48.5f',    None,  -48.5, None,  '-48.5f',   'f, no fps'),

        ('0f',          24,      0,    0,  '0f',       '0f with fps'),
        ('2f',          24,      2, 0.08,  '2f',       '0<f<1s'),
        ('24f',         24,     24,    1,  '1s',       '1s in f'),
        ('48f',         24,     48,    2,  '2s',       '2s in f'),
        ('48.5f',       24,   48.5, 2.02,  '2s 0.5f',  '>1s, fractional'),
        ('-48.5f',      24,  -48.5,-2.02,  '-2s 0.5f', '>1s, fract, -ve'),

        (0,             24,      0,    0,  '0f',       '0 with fps'),
        (1,             24,      1, 0.04,  '1f',       '0<f<1s as int'),
        (2,             24,      2, 0.08,  '2f',       '0<f<1s as int'),
        (-1,            24,    120, 5.00,  '5s',       '-ve as int'),
        (-2,            24,    119, 4.96,  '4s 23f',   '-ve as int'),

          ]

def test_t_examples():

    sif = get_animation('bouncing.sif')

    for example in TESTS:
        try:

            if example[1] is None:
                tag = None
            else:
                sif.tag['fps'] = str(example[1])
                tag = sif.tag

            time = T(example[0], ref=tag)

            message = (
                    "\n"
                    f"              reason: {example[5]}\n"
                    f"Constructor argument: {repr(example[0])}\n"
                    f"            with FPS: {example[1]}\n"
                    )

            if isinstance(example[4], str):
                message += (
                        f"    gives a value of: "
                        f"{example[2]}s {example[3]}f\n"
                        f"which stringifies to: {example[4]}\n"
                        )
            else:
                message += (
                        f"        should raise: {example[4]}\n"
                        )

            assert round(time.frames, 2)==example[2], (
                    f"frames property: {message}"
                    )
            
            if example[3] is not None:
                assert round(time.seconds, 2)==example[3], (
                    f"seconds property: {message}"
                    )

            assert time == example[2],   f"equal to constant: {message}"
            assert time  < example[2]+1, f"less than constant: {message}"
            assert time  > example[2]-1, f"more than constant: {message}"

            t_same   = T(f'{example[2]}f',   ref=tag)
            t_before = T(f'{example[2]-1}f', ref=tag)
            t_after  = T(f'{example[2]+1}f', ref=tag)

            assert time == t_same,   f"equal to another T: {message}"
            assert time != t_before, f"not equal to another T: {message}"
            assert time != t_after,  f"not equal to another T: {message}"
            assert time < t_after,   f"less than another T: {message}"
            assert time > t_before,  f"more than another T: {message}"

            assert hash(time)==hash(t_same),   f"hash equal: {message}"
            assert hash(time)!=hash(t_before), f"hash not equal: {message}"
            assert hash(time)!=hash(t_after),  f"hash not equal: {message}"

            if isinstance(example[4], str):
                assert str(time)==example[4], f"str(): {message}"

            with pytest.raises(AttributeError):
                time.frames = 0

            with pytest.raises(AttributeError):
                time.seconds = 0

            with pytest.raises(AttributeError):
                time.fps = 0

            # (end of the unit tests here)

            assert not isinstance(example[4], Exception), (
                    f"We expected an exception: {example}"
                    )
        except Exception as e:
            if isinstance(e, AssertionError):
                raise

            assert not isinstance(example[4], str), (
                    f"Didn't expect an exception: {example}\n{e}"
                    )
            assert isinstance(e, example[4]), (
                    f"Wrong kind of exception: {example}\n{e}"
                    )

def test_t_ref_types():
    sif = get_animation('bouncing.sif')

    assert sif.fps == 24

    assert T("1s", ref=sif.tag).frames == 24
    assert T("1s", ref=sif).frames == 24

    sif.fps = 50
    assert T("1s", ref=sif.tag).frames == 50
    assert T("1s", ref=sif).frames == 50

    daft = 'This is a string, which should fail'
    with pytest.raises(TypeError):
        assert T("1s", ref=daft)

def test_t_no_params():
    zero_t = T()
    assert zero_t.frames==0
    assert zero_t.seconds==0



from sangfroid.layer import Layer
import sangfroid.value as v
import sangfroid.field as f

@Layer.handles_type()
class Import(Layer):
    SYMBOL = 'I'
    ### {{{
    SYNFIG_VERSION = "0.1"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                        )
    amount               = f.ParamTagField(v.Real, 1.0,
                        )
    blend_method         = f.ParamTagField(v.BlendMethod, v.BlendMethod.COMPOSITE,
                        )
    tl                   = f.ParamTagField(v.X_Y, (-0.5333333611, 0.5333333611),
                        )
    br                   = f.ParamTagField(v.X_Y, (0.5333333611, -0.5333333611),
                        )
    c                    = f.ParamTagField(v.Integer, 1,
                        )
    gamma_adjust         = f.ParamTagField(v.Real, 1.0,
                        )
    filename             = f.ParamTagField(v.String, 'drop.png',
                        )
    time_offset          = f.ParamTagField(v.Time, 0,
                        )

    ### }}}



import pytest
from test import *
from sangfroid.registry import Registry

r = Registry()

@r()
class Wombat():
    pass

@r()
class Tra_La_La():
    pass

@r(name='barney')
class Tahvo:
    pass

def test_registry():

    assert r.from_name('Wombat')==Wombat
    assert r.from_name('wombat')==Wombat
    assert r.from_name('WOMBAT')==Wombat

    assert r.from_name('tralala')==Tra_La_La
    assert r.from_name('tra_la_la')==Tra_La_La
    assert r.from_name('TRALALA')==Tra_La_La
    assert r.from_name('TRA_LA_LA')==Tra_La_La
    assert r.from_name('TrA_lA_La')==Tra_La_La
    assert r.from_name('tra_____la_________la')==Tra_La_La
    assert r.from_name('t_ral_a_la')==Tra_La_La

    assert r.from_name('Barney')==Tahvo
    assert r.from_name('BaRnEy')==Tahvo
    assert r.from_name('BARNEY')==Tahvo
    assert r.from_name('BARNE_____Y')==Tahvo

    with pytest.raises(KeyError):
        r.from_name('fred')

    with pytest.raises(KeyError):
        r.from_name('tahvo')



"""
Many kinds of layer.

The topmost kind of layer in an animation will
always be an instance of `sangfroid.Animation`.
But layers inside that will be instances of one of
these classes.
"""

from sangfroid.layer.layer import *

from sangfroid.layer.blurs import *
from sangfroid.field import *
from sangfroid.layer.geometry import *
from sangfroid.layer.group import *
from sangfroid.layer.other import *
from sangfroid.layer.stylise import *
from sangfroid.layer.text import *
from sangfroid.layer.time import *
from sangfroid.layer.tbd import *
from sangfroid.layer.transform import *

__all__ = [
        cls.__name__.title()
        for cls in Layer.handles_type.handlers.values()
        ]



from sangfroid.layer.layer import Layer
import sangfroid.value as v
import sangfroid.field as f

@Layer.handles_type()
class Text(Layer):
    SYMBOL = '𝕋'

    ### {{{
    SYNFIG_VERSION = "0.5"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                        )
    amount               = f.ParamTagField(v.Real, 1.0,
                        )
    blend_method         = f.ParamTagField(v.BlendMethod, v.BlendMethod.COMPOSITE,
                        )
    text                 = f.ParamTagField(v.String, 'Hello wombat!',
                        )
    color                = f.ParamTagField(v.Color, (1.0, 1.0, 1.0, 1.0),
                        )
    family               = f.ParamTagField(v.String, 'Sans Serif',
                        )
    style                = f.ParamTagField(v.Integer, 0,
                        )
    weight               = f.ParamTagField(v.Integer, 400,
                        )
    direction            = f.ParamTagField(v.Integer, 0,
                        )
    compress             = f.ParamTagField(v.Real, 1.0,
                        )
    vcompress            = f.ParamTagField(v.Real, 1.0,
                        )
    size                 = f.ParamTagField(v.X_Y, (0.25, 0.25),
                        )
    orient               = f.ParamTagField(v.X_Y, (0.5, 0.5),
                        )
    origin               = f.ParamTagField(v.X_Y, (0.0, 0.0),
                        )
    use_kerning          = f.ParamTagField(v.Bool, True,
                        )
    grid_fit             = f.ParamTagField(v.Bool, False,
                        )
    invert               = f.ParamTagField(v.Bool, False,
                        )

    ### }}}



from sangfroid.layer import Layer
import sangfroid.value as v
import sangfroid.field as f
import bs4

@f.ShortcutField.include_from("transformation")
@Layer.handles_type()
class Group(Layer):
    """
    A collection of Layers, which are rendered together.
    """
    SYMBOL = '📂'

    ### {{{
    SYNFIG_VERSION = "0.3"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                                  doc='How deep in the group this layer appears.',
                        )
    amount               = f.ParamTagField(v.Real, 1.0,
                                  doc='1.0 for opaque, 0.0 for transparent.',
                        )
    blend_method         = f.ParamTagField(v.BlendMethod, v.BlendMethod.COMPOSITE,
                                  doc='How this layer will affect the layers beneath it.',
                        )
    origin               = f.ParamTagField(v.X_Y, (0.0, 0.0),
                                  doc='Where the transformation coordinates are measured from.\n\n            You can move this around to move the layer.',
                        )
    transformation       = f.ParamTagField(v.Transformation, {
                                         'offset': (0.0, 0.0),
                                          'angle': 0.0,
                                     'skew_angle': 0.0,
                                          'scale': (1.0, 1.0),
                                        },
                        )
    canvas               = f.ParamTagField(v.Canvas, [],
                                  doc='Any layers which are inside this one.',
                        )
    time_dilation        = f.ParamTagField(v.Real, 1.0,
                        )
    time_offset          = f.ParamTagField(v.Time, 0,
                        )
    children_lock        = f.ParamTagField(v.Bool, False,
                        )
    outline_grow         = f.ParamTagField(v.Real, 0.0,
                        )
    z_range              = f.ParamTagField(v.Bool, False,
                        )
    z_range_position     = f.ParamTagField(v.Real, 0.0,
                        )
    z_range_depth        = f.ParamTagField(v.Real, 0.0,
                        )
    z_range_blur         = f.ParamTagField(v.Real, 0.0,
                        )

    ### }}}

    def _get_children(self,
                 include_descendants = False,
                 ):

        if self._tag.name=='layer':
            canvas = self._tag.find('canvas')
        else:
            canvas = self._tag

        for child in reversed(canvas.contents):
            if not isinstance(child, bs4.element.Tag):
                continue
            if child.name!='layer':
                continue

            result = Layer.from_tag(child)
            yield result

            if include_descendants:
                yield from result.children

    @property
    def children(self):
        yield from self._get_children(
                include_descendants = False,
                )
    @property
    def descendants(self):
        yield from self._get_children(
                include_descendants = True,
                )



from sangfroid.layer import Layer
import sangfroid.value as v
import sangfroid.field as f

@Layer.handles_type()
class Scale(Layer):
    SYMBOL = '⚖️' # yeah, a bit contrived

    ### {{{
    PARAMS = {
        "amount": v.Real,
        "center": v.Vector,
    }












    ### }}}

@Layer.handles_type()
class Zoom(Scale):
    ### {{{
    SYNFIG_VERSION = "0.1"

    amount               = f.ParamTagField(v.Real, 0.0,
                        )
    center               = f.ParamTagField(v.X_Y, (0.0, 0.0),
                        )

    ### }}}
    pass # XXX do they differ?

@Layer.handles_type()
class Translate(Layer):
    SYMBOL = '⇄'

    ### {{{
    SYNFIG_VERSION = "0.1"

    origin               = f.ParamTagField(v.X_Y, (0.0, 0.0),
                        )

    ### }}}

@Layer.handles_type()
class Rotate(Layer):
    SYMBOL = '🗘'

    ### {{{
    SYNFIG_VERSION = "0.1"

    origin               = f.ParamTagField(v.X_Y, (0.0, 0.0),
                        )
    amount               = f.ParamTagField(v.Angle, 0.0,
                        )

    ### }}}



import sangfroid
from bywords.bywords import Bywords
import logging

logger = logging.getLogger('bywords')

def set_constant_waypoint(tl, time, value):
    s = sangfroid.value.String(value)
    t = sangfroid.T(
            f"{time}s",
            ref = tl.parent.tag,
            )
    waypoint = sangfroid.value.Waypoint(
            time = t,
            value = s,
            before = 'constant',
            after = 'constant',
            )
    tl += waypoint

def adjust(subtitles, animation,
           letter_at_a_time = False,
           ):
    bw = Bywords(
            sif=animation,
            srt=subtitles,
            letter_at_a_time = letter_at_a_time,
            )

    if logger.level<logging.WARNING:
        bw.dump_findings_to_logger()

    for obj in set([o.item for o in bw]):
        obj.is_animated = True
        timeline = obj['text'].timeline

        set_constant_waypoint(timeline, 0, '')

        for word in bw:
            if word.item!=obj:
                continue

            seconds = f"{word.time/1000}s"

            set_constant_waypoint(timeline,
                                  word.time,
                                  word.text,
                                  )



from sangfroid.layer import Layer
import sangfroid.value as v
import sangfroid.field as f

@Layer.handles_type()
class Star(Layer):
    SYMBOL = '⭐'

    ### {{{
    SYNFIG_VERSION = "0.1"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                        )
    amount               = f.ParamTagField(v.Real, 1.0,
                        )
    blend_method         = f.ParamTagField(v.BlendMethod, v.BlendMethod.COMPOSITE,
                        )
    color                = f.ParamTagField(v.Color, (1.0, 1.0, 1.0, 1.0),
                        )
    origin               = f.ParamTagField(v.X_Y, (0.0, 0.0),
                        )
    invert               = f.ParamTagField(v.Bool, False,
                        )
    antialias            = f.ParamTagField(v.Bool, True,
                        )
    feather              = f.ParamTagField(v.Real, 0.0,
                        )
    blurtype             = f.ParamTagField(v.Integer, 1,
                        )
    winding_style        = f.ParamTagField(v.Integer, 0,
                        )
    radius1              = f.ParamTagField(v.Real, 1.0,
                        )
    radius2              = f.ParamTagField(v.Real, 0.38,
                        )
    angle                = f.ParamTagField(v.Angle, 90.0,
                        )
    points               = f.ParamTagField(v.Integer, 5,
                        )
    regular_polygon      = f.ParamTagField(v.Bool, False,
                        )

    ### }}}

@Layer.handles_type()
class Solid_Color(Layer):
    SYMBOL = '▊'

    ### {{{
    SYNFIG_VERSION = "0.1"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                        )
    amount               = f.ParamTagField(v.Real, 1.0,
                        )
    blend_method         = f.ParamTagField(v.BlendMethod, v.BlendMethod.COMPOSITE,
                        )
    color                = f.ParamTagField(v.Color, (1.0, 1.0, 1.0, 1.0),
                        )

    ### }}}

@Layer.handles_type()
class Region(Layer):
    SYMBOL = '🟤'

    ### {{{
    SYNFIG_VERSION = "0.1"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                        )
    amount               = f.ParamTagField(v.Real, 1.0,
                        )
    blend_method         = f.ParamTagField(v.BlendMethod, v.BlendMethod.COMPOSITE,
                        )
    color                = f.ParamTagField(v.Color, (1.0, 1.0, 1.0, 1.0),
                        )
    origin               = f.ParamTagField(v.X_Y, (0.0, 0.0),
                        )
    invert               = f.ParamTagField(v.Bool, False,
                        )
    antialias            = f.ParamTagField(v.Bool, True,
                        )
    feather              = f.ParamTagField(v.Real, 0.0,
                        )
    blurtype             = f.ParamTagField(v.Integer, 1,
                        )
    winding_style        = f.ParamTagField(v.Integer, 0,
                        )
    bline                = f.NotImplementedField("Bline",
                        )

    ### }}}

@Layer.handles_type()
class Rectangle(Layer):
    SYMBOL = '🟦'

    ### {{{
    SYNFIG_VERSION = "0.2"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                        )
    amount               = f.ParamTagField(v.Real, 1.0,
                        )
    blend_method         = f.ParamTagField(v.BlendMethod, v.BlendMethod.COMPOSITE,
                        )
    color                = f.ParamTagField(v.Color, (1.0, 1.0, 1.0, 1.0),
                        )
    point1               = f.ParamTagField(v.X_Y, (0.0, 0.0),
                        )
    point2               = f.ParamTagField(v.X_Y, (1.0, 1.0),
                        )
    expand               = f.ParamTagField(v.Real, 0.0,
                        )
    invert               = f.ParamTagField(v.Bool, False,
                        )
    feather_x            = f.ParamTagField(v.Real, 0.0,
                        )
    feather_y            = f.ParamTagField(v.Real, 0.0,
                        )
    bevel                = f.ParamTagField(v.Real, 0.0,
                        )
    bevCircle            = f.ParamTagField(v.Bool, True,
                        )

    ### }}}

@Layer.handles_type()
class Polygon(Layer):
    SYMBOL = '⭓'

    ### {{{
    SYNFIG_VERSION = "0.1"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                        )
    amount               = f.ParamTagField(v.Real, 1.0,
                        )
    blend_method         = f.ParamTagField(v.BlendMethod, v.BlendMethod.COMPOSITE,
                        )
    color                = f.ParamTagField(v.Color, (1.0, 1.0, 1.0, 1.0),
                        )
    origin               = f.ParamTagField(v.X_Y, (0.0, 0.0),
                        )
    invert               = f.ParamTagField(v.Bool, False,
                        )
    antialias            = f.ParamTagField(v.Bool, True,
                        )
    feather              = f.ParamTagField(v.Real, 0.0,
                        )
    blurtype             = f.ParamTagField(v.Integer, 1,
                        )
    winding_style        = f.ParamTagField(v.Integer, 0,
                        )
    vector_list          = f.NotImplementedField("Dynamic_List",
                        )

    ### }}}

@Layer.handles_type()
class Outline(Layer):
    SYMBOL = '⭔'

    ### {{{
    SYNFIG_VERSION = "0.3"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                        )
    amount               = f.ParamTagField(v.Real, 1.0,
                        )
    blend_method         = f.ParamTagField(v.BlendMethod, v.BlendMethod.COMPOSITE,
                        )
    color                = f.ParamTagField(v.Color, (0.0, 0.0, 0.0, 1.0),
                        )
    origin               = f.ParamTagField(v.X_Y, (0.0, 0.0),
                        )
    invert               = f.ParamTagField(v.Bool, False,
                        )
    antialias            = f.ParamTagField(v.Bool, True,
                        )
    feather              = f.ParamTagField(v.Real, 0.0,
                        )
    blurtype             = f.ParamTagField(v.Integer, 1,
                        )
    winding_style        = f.ParamTagField(v.Integer, 0,
                        )
    bline                = f.NotImplementedField("Bline",
                        )
    width                = f.ParamTagField(v.Real, 0.0166666667,
                        )
    expand               = f.ParamTagField(v.Real, 0.0,
                        )
    sharp_cusps          = f.ParamTagField(v.Bool, True,
                        )
    round_tip            = f.ParamArrayField(v.Bool, True,
                        )
    homogeneous_width    = f.ParamTagField(v.Bool, True,
                        )

    ### }}}

@Layer.handles_type()
class Circle(Layer):
    SYMBOL = '🔵'

    ### {{{
    SYNFIG_VERSION = "0.2"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                        )
    amount               = f.ParamTagField(v.Real, 1.0,
                        )
    blend_method         = f.ParamTagField(v.BlendMethod, v.BlendMethod.COMPOSITE,
                        )
    color                = f.ParamTagField(v.Color, (1.0, 1.0, 1.0, 1.0),
                        )
    radius               = f.ParamTagField(v.Real, 1.0,
                        )
    feather              = f.ParamTagField(v.Real, 0.0,
                        )
    origin               = f.ParamTagField(v.X_Y, (0.0, 0.0),
                        )
    invert               = f.ParamTagField(v.Bool, False,
                        )

    ### }}}

@Layer.handles_type()
class Checker_Board(Layer):
    SYMBOL = '🙾'

    ### {{{
    SYNFIG_VERSION = "0.2"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                        )
    amount               = f.ParamTagField(v.Real, 1.0,
                        )
    blend_method         = f.ParamTagField(v.BlendMethod, v.BlendMethod.COMPOSITE,
                        )
    color                = f.ParamTagField(v.Color, (1.0, 1.0, 1.0, 1.0),
                        )
    origin               = f.ParamTagField(v.X_Y, (0.125, 0.125),
                        )
    size                 = f.ParamTagField(v.X_Y, (0.25, 0.25),
                        )
    antialias            = f.ParamTagField(v.Bool, True,
                        )

    ### }}}

@Layer.handles_type()
class Advanced_Outline(Layer):
    SYMBOL = '⬡'

    ### {{{
    SYNFIG_VERSION = "0.3"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                        )
    amount               = f.ParamTagField(v.Real, 1.0,
                        )
    blend_method         = f.ParamTagField(v.BlendMethod, v.BlendMethod.COMPOSITE,
                        )
    color                = f.ParamTagField(v.Color, (0.0, 0.0, 0.0, 1.0),
                        )
    origin               = f.ParamTagField(v.X_Y, (0.0, 0.0),
                        )
    invert               = f.ParamTagField(v.Bool, False,
                        )
    antialias            = f.ParamTagField(v.Bool, True,
                        )
    feather              = f.ParamTagField(v.Real, 0.0,
                        )
    blurtype             = f.ParamTagField(v.Integer, 1,
                        )
    winding_style        = f.ParamTagField(v.Integer, 0,
                        )
    bline                = f.NotImplementedField("Bline",
                        )
    width                = f.ParamTagField(v.Real, 0.0166666667,
                        )
    expand               = f.ParamTagField(v.Real, 0.0,
                        )
    start_tip            = f.ParamTagField(v.Integer, 1,
                        )
    end_tip              = f.ParamTagField(v.Integer, 1,
                        )
    cusp_type            = f.ParamTagField(v.Integer, 0,
                        )
    smoothness           = f.ParamTagField(v.Real, 1.0,
                        )
    homogeneous          = f.ParamTagField(v.Bool, True,
                        )
    wplist               = f.NotImplementedField("Wplist",
                        )
    dash_enabled         = f.ParamTagField(v.Bool, False,
                        )
    dilist               = f.NotImplementedField("Dilist",
                        )
    dash_offset          = f.ParamTagField(v.Real, 0.0,
                        )

    ### }}}



def normalise_synfig_layer_type_name(s):
    """
    Changes a value of the "name" attribuyte on a <layer> tag
    into its normal form.

    Args:
        s (str): the name

    Returns:
        str
    """
    return s.lower().replace('_', '')

def type_and_value_to_str(t, v):
    if v is None:
        return None
    elif issubclass(t, bool):
        return str(bool(v)).lower()
    else:
        return str(v)

def type_and_str_to_value(t, s):
    if v is None:
        return None
    elif issubclass(t, bool):
        return v.lower()=='true'
    else:
        return t(v)



from enum import Enum
from sangfroid.value.simple import Integer

class BlendMethod(Enum):
    ### {{{

    COMPOSITE = 0
    STRAIGHT = 1
    BRIGHTEN = 2
    DARKEN = 3
    ADD = 4
    SUBTRACT = 5
    MULTIPLY = 6
    DIVIDE = 7
    COLOR = 8
    HUE = 9
    SATURATION = 10
    LUMINANCE = 11
    BEHIND = 12
    ONTO = 13
    ALPHA_BRIGHTEN = 14
    ALPHA_DARKEN = 15
    SCREEN = 16
    HARD_LIGHT = 17
    DIFFERENCE = 18
    ALPHA_OVER = 19
    OVERLAY = 20
    STRAIGHT_ONTO = 21
    ADD_COMPOSITE = 22
    ALPHA = 23
    
    ### }}}

    @classmethod
    def from_tag(cls, t):
        v = Integer.from_tag(t)
        return cls(v.value)

    def as_python_expression(self):
        return 'v.'+str(self)



from sangfroid.layer import Layer
import sangfroid.value as v
import sangfroid.field as f

@Layer.handles_type()
class Shade(Layer):
    SYMBOL = '👓'

    ### {{{
    SYNFIG_VERSION = "0.2"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                        )
    amount               = f.ParamTagField(v.Real, 0.75,
                        )
    blend_method         = f.ParamTagField(v.BlendMethod, v.BlendMethod.BEHIND,
                        )
    color                = f.ParamTagField(v.Color, (1.0, 1.0, 1.0, 1.0),
                        )
    origin               = f.ParamTagField(v.X_Y, (0.200000003, -0.200000003),
                        )
    size                 = f.ParamTagField(v.X_Y, (0.1000000015, 0.1000000015),
                        )
    type_                = f.ParamTagField(v.Integer, 1,
                        )
    invert               = f.ParamTagField(v.Bool, False,
                        )

    ### }}}

@Layer.handles_type()
class Bevel(Layer):
    SYMBOL = '🫴'

    ### {{{
    SYNFIG_VERSION = "0.2"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                        )
    amount               = f.ParamTagField(v.Real, 0.75,
                        )
    blend_method         = f.ParamTagField(v.BlendMethod, v.BlendMethod.ONTO,
                        )
    type_                = f.ParamTagField(v.Integer, 1,
                        )
    color1               = f.ParamTagField(v.Color, (1.0, 1.0, 1.0, 1.0),
                        )
    color2               = f.ParamTagField(v.Color, (0.0, 0.0, 0.0, 1.0),
                        )
    angle                = f.ParamTagField(v.Angle, 135.0,
                        )
    depth                = f.ParamTagField(v.Real, 0.2,
                        )
    softness             = f.ParamTagField(v.Real, 0.1,
                        )
    use_luma             = f.ParamTagField(v.Bool, False,
                        )
    solid                = f.ParamTagField(v.Bool, False,
                        )
    fake_origin          = f.ParamTagField(v.X_Y, (0.0, 0.0),
                        )

    ### }}}



from bs4 import BeautifulSoup
import sangfroid
import os
import glob
import re
import collections
import math
import re
import logging
import functools

logger = logging.getLogger('bywords')
LOGGING_FORMAT = '%(time)10s %(layer)10s %(str)s'

WHITESPACE = re.compile('([ \t\n]+)')
NOTHING = re.compile('()')

NON_WORD_LETTERS = re.compile('[^a-z0-9 ]') # FIXME not very i18n-friendly

PREFIX = 'bywords-'

bywords_groups = None

@functools.total_ordering
class WordTiming:
    def __init__(self, text, item):
        self.text = text
        self.item = item
        self.time = 0

    def __lt__(self, other):
        return self.time<other.time

    def __eq__(self, other):
        return self.time==other.time

    def __str__(self):
        return LOGGING_FORMAT % {
            'time': '%0.02g' % (self.time,),
            'layer': self.item,
            'str': self.text,
            }

    __repr__ = __str__

class BywordsGroup:
    def __init__(self, group):
        self.group = group

        self.texts = [
                (self.tidy_string(item.desc), item,)
                for item in
                list(reversed(group.find_all('text')))
                ]

        self.full_text = ' '.join(
                [s[0] for s in self.texts])

    @classmethod
    def tidy_string(self, n):
        return NON_WORD_LETTERS.sub('', n.lower())

    def matches(self, s, duration, start=0,
                letter_at_a_time = False,
                ):

        word_timings = []

        if letter_at_a_time:
            splitting_re = NOTHING
        else:
            splitting_re = WHITESPACE

        if s!=self.full_text:
            return None

        for _, item in self.texts:
            item['text'].is_animated = False

            words = [w for w in splitting_re.split(item.desc)
                     if w!=''
                     ]

            i = 0
            while words:
                word = words.pop(0)

                i += len(word)

                word_timings.append(
                        WordTiming(
                            text = item.desc[:i],
                            item = item,
                            ))

                if words and words[0].strip()=='':
                    i += len(words.pop(0))

            s = s[len(item.desc):]

        total_letters = sum([len(w.text) for w in word_timings])
        ms_per_letter = duration/total_letters

        time = start
        for w in word_timings:
            w.time = time/1000
            time += len(w.text)*ms_per_letter

        return word_timings

    def __str__(self):
        return self.group.desc

    __repr__ = __str__

    def __lt__(self, other):
        return str(self) < str(other)

    def __eq__(self, other):
        return str(self) < str(other)

class Bywords:
    def __init__(self, srt, sif,
                 letter_at_a_time = False,
                 ):
        self.srt = srt
        self.sif = sif

        self.no_matches = []

        self.lines = [
                (
                    line.start.ordinal,
                    line.duration.ordinal,
                    BywordsGroup.tidy_string(line.text),
                    )
                for line in srt
                ]

        self.groups = sorted([
            BywordsGroup(group)
            for group in self.sif.find_all(sangfroid.layer.Group)
            if group.desc is not None
            and group.desc.lower().startswith('bywords')
            ])

        result = {}
        found = None

        for start, duration, text in self.lines:
            for group in self.groups:
                found = group.matches(text,
                                      duration,
                                      start,
                                      letter_at_a_time,
                                 )
                if found:
                    break
            if found:
                for word in found:
                    result[word.time] = word
            else:
                self.no_matches.append(text)

        self.words = sorted(list(result.values()))

    def dump_findings_to_logger(self):
        logger.info("\n\n=== Source timings ===\n")

        logger.info("---  %12s %s", "Time (s)", "Normalised text")
        for start, duration, text in self.lines:
            logger.info("    %12s %s", start/1000, text)

        logger.info("\n\n=== Source layers ===\n")

        logger.info("---  Normalised text")
        for group in self.groups:
            logger.info("   %s", group.full_text)

        logger.info("\n\n=== Matches ===\n")

        logger.info(LOGGING_FORMAT % {
            'time': 'At time (s)',
            'layer': 'on layer',
            'str': 'set string',
            })
        if self.words:
            for word_timing in self.words:
                logger.info("    %s", word_timing)

        else:
            logger.info("There were no matches.")

        logger.info("\n\n=== Unmatched ===\n")

        if self.no_matches:
            for line in self.no_matches:
                logger.info("    %s", line)
        else:
            logger.info("Everything was matched.")

    def __iter__(self):
        return iter(self.words)



import copy
import bs4
import functools
import copy
from sangfroid.registry import Registry
from sangfroid.t import T

class Value:

    ANIMATED = 'animated'

    def __init__(self, *args):

        if len(args)==1 and isinstance(args[0], bs4.element.Tag):
            self._tag = args[0]
        else:
            self._tag = self._get_empty_tag()

            if len(args)==1:
                self.value = args[0]
            else:
                self.value = args

        assert self._tag is not None

    @classmethod
    def _get_empty_tag(cls, name=None):
        name = name or cls.__name__.lower()
        result = bs4.Tag(name=name)
        return result

    @property
    def tag(self):
        return self._tag

    @property
    def is_animated(self):
        """
        Whether the value is animated.

        Returns:
            bool
        """
        return self._tag.name==self.ANIMATED

    @is_animated.setter
    def is_animated(self, v):
        self._set_animated(v)

    def _set_animated(self,
                      whether,
                      adjust_contents = True,
                      ):
        whether = bool(whether)
        if whether==self.is_animated:
            pass
        elif whether:

            if adjust_contents:
                former_value = self.value

            our_type = self._tag.name
            self._tag.attrs = {}
            self._tag.name = self.ANIMATED
            self._tag['type'] = our_type

            if adjust_contents:
                self._tag.clear()
                self.timeline[0] = former_value
        else:

            if adjust_contents:
                timeline = self.timeline
                first_value = timeline.values()[0]
            else:
                first_value = None

            self._tag.clear()

            new_tag = self._get_empty_tag()
            self._tag.replace_with(new_tag)
            self._tag = new_tag

            if first_value is not None:
                old_tag = self._tag

                self._tag = copy.deepcopy(first_value.value._tag)
                if old_tag.parent is not None:
                    old_tag.replace_with(self._tag)

    @property
    def timeline(self):
        result = Timeline.__new__(Timeline)
        result.parent = self
        return result

    @timeline.setter
    def timeline(self, v):

        if isinstance(v, list) and all([n for n in v if isinstance(n, Waypoint)]):
            self._set_waypoints(v)
        elif isinstance(v, dict):
            self.timeline = list(v.values())
        elif isinstance(v, Timeline):
            if v.parent is self:
                return
            self._set_waypoints(v.values())
        else:
            raise TypeError("A timeline can only be set to another timeline or "
                            "a dict or list of Waypoints.")

    def _waypoint_tags(self):
        """
        A list of Beautiful Soup tags of waypoints on our timeline.

        The list is in the same order it appears in the file.
        If we're not animated, the list will have no members.

        Returns:
            list of Tag
        """

        if self._tag.name!=Value.ANIMATED:
            return []

        result = [wt for wt in self._tag
                  if isinstance(wt, bs4.element.Tag)]

        return result

    def _waypoints(self):
        """
        A dict of Waypoints on our timeline.

        If we're not animated, the dict will have no members.

        Returns:
            dict mapping T to Waypoint
        """

        waypoints = self._waypoint_tags()

        if not waypoints:
            return {}

        our_type = self._tag['type']

        values = [Waypoint.from_tag(wt,
                                    our_type = our_type,
                                    )
                  for wt in waypoints]

        result = dict([(v.time, v) for v in values])

        return result

    def _set_waypoints(self, v):
        """
        Sets our timeline to contain exactly the given sequence of waypoints.

        They will be stored sorted, with newlines between them.

        This necessarily involves setting whether we're animated.

        Args:
            v (list of Waypoints): the waypoints.
        """

        if not v:
            self._set_animated(
                    whether = False,
                    )
            return

        self._set_animated(
                whether = True,
                adjust_contents = False,
                )
        self._tag.clear()

        for i, w in enumerate(sorted(v)):
            if i!=0:
                self._tag.append('\n')
            self._tag.append(w.tag)

    def __len__(self):
        return len(self._waypoints())

    @property
    def our_type(self):
        """
        The name of the Synfig layer type.

        For example, 'circle' or 'group'.

        Returns:
            str
        """
        result = self._tag.name
        if result==Value.ANIMATED:
            result = self._tag['type']

        return result

    def _str_inner(self):
        return str(self.value)

    def __str__(self):
        if self.is_animated:
            return '(animated)'
        else:
            return self._str_inner()

    def __repr__(self):
        return '['+self.__class__.__name__+' '+str(self)+']'

    @classmethod
    def _subfields(cls):
        """
        Returns a set of names of keys
        generally found within values of this class.

        Used by sangfroid.layer.include_shortcuts().

        Returns:
            set
        """
        return set()

    @property
    def value(self):
        raise NotImplementedError()

    def __eq__(self, other):
        if isinstance(other, Value):
            return self.value == other.value

        return self.value == other

    def as_python_expression(self):
        """
        A Python expression which could be passed to the constructor
        of this class in order to recreate this value.

        Used by `etc/pick-and-mix-to-layers.py`.
        """
        return str(self)

    ########################

    # Factories, and setup for factories

    handles_type = Registry()

    @classmethod
    def from_tag(cls, tag,
                 ):

        if tag.name==cls.ANIMATED:

            type_name = tag['type']
            if type_name is None:
                raise ValueError(f"Animated values need a type: {tag}")

        else:
            type_name = tag.name

        result_type = cls.handles_type.from_name(name=type_name)
        result = result_type._construct_from(tag)

        return result

    @classmethod
    def _construct_from(cls, tag):
        return cls(tag)

#######################

class Timeline:
    r"""
    A wrapper for a Value, giving access to its Waypoints.

    This can't be created directly; it should only be created by
    a Value. It holds no state of its own, other than the reference
    to the Value which created it.

    Fields:
        parent (Value): the Value which created this Timeline
    """

    def __init__(self):
        raise NotImplementedError(
                "Don't construct timelines directly."
                )

    def __iter__(self):
        for t,w in sorted(self.parent._waypoints().items()):
            yield w

    def _ensure_fps(self, t):
        if isinstance(t, (int, float, str)):
            return T(t, ref = self.parent._tag)
        elif isinstance(t, T):
            if t._fps is None:
                return T(t._frames, self.parent._tag)
            else:
                return t
        else:
            raise TypeError("I need T, or an int, float, str to create a T. "
                    f"You gave me {type(t)}.")

    def keys(self):
        return list(self.parent._waypoints().keys())

    def values(self):
        return list(self.parent._waypoints().values())

    def items(self):
        return list(self.parent._waypoints().items())

    def __iadd__(self, waypoints):
        self.add(waypoints,
            overwrite = True,
                 )

    def add(self, waypoints,
            overwrite = False,
            ):

        def raise_argument_error():
            raise TypeError(
                    "The argument to add() "
                    "must be either a Waypoint or a list of Waypoints.")

        if isinstance(waypoints, Waypoint):
            waypoints = [waypoints]
        elif isinstance(waypoints, list):
            pass
        else:
            raise_argument_error()

        # check they're sensible
        for w in waypoints:
            if not isinstance(w, Waypoint):
                raise_argument_error()
            elif not isinstance(w.tag, bs4.Tag):
                raise TypeError(
                        f'{w} has a tag of type {type(w.tag)}')

        existing = self.parent._waypoints()
        clashes = [
            (old.time, new.time)
                for old in existing.values()
                for new in waypoints
                if old.time==new.time
                ]

        if overwrite:
            for old, _ in clashes:
                del existing[old]
        elif clashes:
            raise ValueError("There are already Waypoints with those "
                             "times in this timeline:\n"
                             f"{clashes}")

        self.parent.is_animated = True

        existing |= dict([(w.time, w) for w in waypoints])

        self.parent.tag.clear()

        for w in sorted(existing.values()):
            self.parent.tag.append(w.tag)

        return self

    def __getitem__(self, time):
        if 'x_y' in str(self.parent._tag):
            raise ValueError()
        for t, wt in self.parent._waypoints().items():
            if t==time:
                return wt
            elif t>time:
                raise KeyError(time)

        raise KeyError(time)

    def __setitem__(self, t, v):

        t = self._ensure_fps(t)

        if isinstance(v, Waypoint):
            new_waypoint = v
            new_waypoint.time = t
        elif isinstance(v, self.parent.__class__):
            new_waypoint = Waypoint(
                    time = t,
                    value = v,
                    ref = self.parent.tag,
                    )
        else:
            new_waypoint = Waypoint(
                    time = t,
                    value = self.parent.__class__(v),
                    ref = self.parent.tag,
                    )

        if not self.parent.is_animated:
            self.parent._tag.clear()
            self.parent._set_animated(
                    whether = True,
                    adjust_contents = False,
                    )

        waypoints = self.parent._waypoints()

        waypoints[t] = new_waypoint

        self.parent._set_waypoints(waypoints.values())

    def __delitem__(self, t):

        if isinstance(t, (int, float, str)):
            t = T(t, ref = self.parent._tag)
        elif isinstance(t, T):
            t = self._ensure_fps(t)

        waypoints = self.parent._waypoints()

        waypoints.__delitem__(t)

        self.parent._set_waypoints(waypoints.values())

    def __eq__(self, other):
        if isinstance(other, Timeline):
            return self.values()==other.values()
        else:
            return self.values()==list(other)

    def __str__(self):
        result = (
                f'[timeline of {self.parent.__class__.__name__}:'
                f'{self.parent._waypoints()}]'
                )
        return result

    def __len__(self):
        return len(self.keys())

    def __bool__(self):
        return len(self)!=0

    __repr__ = __str__

#######################

INTERPOLATION_TYPES = {
        # UI name    SVG name   emoji
        'tcb':      ('auto',     '🟢'),
        'clamped':  ('clamped',  '🔶'),
        'constant': ('constant', '🟥'),
        'linear':   ('linear',   '🌽'), # yeah, I know it's sweetcorn
        'ease':     ('halt',     '🫐'), # blueberry
        'undefined': (None,      '🪨'), # rock
        }

INTERPOLATION_TYPE_SYNONYMS = dict(
        [(v[0], k)
         for k,v in INTERPOLATION_TYPES.items()
         if v[0] is not None])

@functools.total_ordering
class Waypoint:
    def __init__(self, time, value, before='clamped', after='clamped',
                 ref = None,
                 ):

        if not isinstance(value, Value):
            raise TypeError(value)

        if value.is_animated:
            raise ValueError("Waypoints can't have animated values")

        if isinstance(time, T):
            self.time = time
        elif isinstance(time, (int, float, str)):
            self.time = T(time,
                          ref = ref,
                          )
        else:
            raise TypeError(
                    f"time parameter should be T, or numeric, or str: {time}")

        self._before = self._check_interpolation_type(before, True)
        self._after = self._check_interpolation_type(after, True)
        self.value = value

    @property
    def before(self):
        return self._before

    @before.setter
    def before(self, v):
        self._before = self._check_interpolation_type(v, False)

    @property
    def after(self):
        return self._after

    @after.setter
    def after(self, v):
        self._after = self._check_interpolation_type(v, False)

    @classmethod
    def _check_interpolation_type(cls, v, from_constructor):

        if v=='undefined':
            if from_constructor:
                raise ValueError(
                        "Waypoints can't have interpolations "
                        "of 'undefined'."
                        )
            else:
                raise ValueError(
                        "You can't set waypoint interpolations "
                        "to 'undefined'."
                        )

        if v in INTERPOLATION_TYPES:
            return v

        if v in INTERPOLATION_TYPE_SYNONYMS:
            return INTERPOLATION_TYPE_SYNONYMS[v]

        raise ValueError(f"Unknown interpolation type: {v}")

    @property
    def tag(self):
        result = bs4.Tag(name="waypoint")
        result['time'] = self.time
        result['before'] = self._before
        result['after'] = self._after
        result.append(
                copy.copy(
                    self.value.tag
                    )
                )
        return result

    @classmethod
    def from_tag(cls, tag,
                 our_type = None,
                 ):
        if tag.name!='waypoint':
            raise ValueError("Waypoints must be called <waypoint>: "
                                f"{tag}")

        try:
            time = T(tag['time'], ref=tag)
        except ValueError:
            raise ValueError(
                    "If a value isn't attached to a document, "
                    "T-values in its timeline must be"
                    "expressed in frames: {tag}"
                    )

        v = [t for t in tag.children if isinstance(t, bs4.element.Tag)]

        if len(v)==0:
            raise ValueError(f"Waypoint without a value: {w}")
        elif len(v)!=1:
            raise ValueError(
                    f"Waypoint with multiple values: {w}")
        elif v[0].name==Value.ANIMATED:
            raise ValueError("Values in waypoints cannot themselves "
                             "be animated")
        elif our_type is not None and v[0].name!=our_type:
            raise ValueError(
                    "Waypoint type must match parent: "
                    f"parent={our_type}, child={v[0].name}")

        result = cls(
                time = time,
                value = Value.from_tag(v[0]),
                before = tag['before'],
                after = tag['after'],
                )

        return result

    def __lt__(self, other):
        return self.time < other.time

    def __eq__(self, other):
        if not isinstance(other, Waypoint):
            return False

        return self.time == other.time

    def __str__(self):
        return '[%3s ' % (self.time,) + (
                f'{INTERPOLATION_TYPES[self.before][1]}-'
                f'{INTERPOLATION_TYPES[self.after][1]} - '
                f'{self.value}]'
                )

    __repr__ = __str__

__all__ = [
        'Value',
        'Waypoint',
        ]



import gzip
import os
from bs4 import BeautifulSoup

class Format:
    """
    A disk file, containing an animation saved in a particular format.

    It also allows access to files dependent on the main file.

    Fields:
        filename (str): the filename.

    Class fields:
        extensions (tuple of str): the filename extensions which
            can be used for this format, without leading dots.
        magic_number (bytes): the first two characters of
            files of this format.
    """

    def __init__(self):
        raise NotImplementedError(
                "Create formats using from_filename().")

    def main_file(self):
        """
        Returns the parsed XML of the main document.

        For the filename, see the `filename` field.

        Returns:
            BeautifulSoup.Tag
        """
        raise NotImplementedError()

    def __getitem__(self, v):
        """
        Looks up a file referred to by the main file.
        """
        return self._subfiles[v]

    def __len__(self):
        """
        Returns the number of files referred to in the main file.

        Returns:
            int
        """
        return len(self._subfiles)

    def items(self):
        return self._subfiles.items()

    def keys(self):
        return self._subfiles.keys()

    def values(self):
        return self._subfiles.values()

    @classmethod
    def from_filename(cls, filename,
                      load = True,
                      ):
        """
        Given a filename, returns a Format of the appropriate subclass,
        possibly loaded from the file of that name.

        Args:
            filename (str): the filename
            load (bool): if True, load the file with that name

        Returns:
            Format

        Raises:
            ValueError: if the filename isn't in a format we know
                how to handle.
            FileNotFoundError: if load==True and the file doesn't
                exist.
        """
        assert filename is not None

        if load:
            with open(filename, 'rb') as f:
                magic = f.read(2)

            if not magic:
                raise ValueError("The file {filename} is empty.")

            handlers = [h for h in cls.handlers()
                        if h.magic_number==magic]
            if not handlers:
                raise ValueError(
                        f"The file {filename} isn't in a format I know how "
                        "to handle.")
        else:
            extension = os.path.splitext(filename)[1].lower()
            if extension.startswith('.'):
                extension = extension[1:]

            handlers = [h for h in cls.handlers()
                        if extension in h.extensions]
            if not handlers:
                raise ValueError(
                        "I don't know how to handle files with the "
                        f"extension {extension}."
                        )

        assert len(handlers)==1

        handler = handlers[0]

        result = handler.__new__(handler)
        result.filename = filename
        return result

    def save(self, content, filename=None):
        """
        Saves a file to disk.

        Args:
            content (bs4.Document): the XML document to save
            filename (str or None): the filename to save under;
                if None, we use self.filename; if not None,
                this is a "save as", so self.filename will
                be set to this value.
        """
        raise NotImplementedError()

    def _filename_for_saving(self, filename):
        if filename is None:
            filename = self.filename
        else:
            self.filename = filename

        return filename

    def _write_to_file(self, f, content):
        f.write(str(content).encode('UTF-8'))

    @classmethod
    def handlers(cls):
        """
        Returns all known subclasses of Format.
        """
        return [
                h for h in globals().values()
                if isinstance(h, type)
                and issubclass(h, Format)
                and h!=Format
                ]

class FileContextHandler:
    def __init__(self, f):
        self.soup = BeautifulSoup(
                f,
                features = 'xml',
                )

    def __enter__(self):
        return self.soup

    def __exit__(self, exc_type, exc_value, traceback):
        pass

class Sif(Format):

    magic_number = b'<?' # start of XML doctype
    extensions = ('sif',)

    def main_file(self):
        return FileContextHandler(open(self.filename, 'r'))

    def save(self, content, filename=None):
        filename = self._filename_for_saving(filename)

        with open(filename, 'wb') as f:
            self._write_to_file(f, content)

class Sifz(Format):

    magic_number = b'\x1f\x8b' # gzip header
    extensions = ('sifz',)

    def main_file(self):
        return FileContextHandler(gzip.open(self.filename, 'r'))

    def save(self, content, filename=None):
        filename = self._filename_for_saving(filename)

        with gzip.open(filename, 'wb') as f:
            self._write_to_file(f, content)

class Sfg(Format):
    
    magic_number = b'PK' # zipfile header (RIP Phil Katz)
    extensions = ('sfg',)

    def main_file(self):
        raise ValueError(
                '.sfg is not yet supported. See: \n'
                'https://gitlab.com/marnanel/sangfroid/-/issues/2'
                )

class Blank(Format):
    """
    The blank animation you get if you instantiate Animation without
    a filename.

    Unlike most subclasses of Format, this can be instantiated
    directly.
    """

    magic_number = None
    extensions = tuple()

    def __init__(self):
        self.filename = None

    def main_file(self):
        return FileContextHandler(BLANK_ANIMATION)

    def save(self, *args, **kwargs):
        raise ValueError(
                'It makes no sense to save to new'
                )

BLANK_ANIMATION = """<?xml version="1.0" encoding="UTF-8"?>
<canvas
        version="1.2"
        width="480" height="270"
        xres="2834.645669" yres="2834.645669"
        gamma-r="1.000000" gamma-g="1.000000" gamma-b="1.000000"
        view-box="-4.000000 2.250000 4.000000 -2.250000"
        antialias="1" fps="24.000"
        begin-time="0f" end-time="5s"
        bgcolor="0.500000 0.500000 0.500000 1.000000"
        >
  <name>New animation</name>
  <meta name="background_first_color" content="0.880000 0.880000 0.880000"/>
  <meta name="background_rendering" content="0"/>
  <meta name="background_second_color" content="0.650000 0.650000 0.650000"/>
  <meta name="background_size" content="15.000000 15.000000"/>
  <meta name="grid_color" content="0.623529 0.623529 0.623529"/>
  <meta name="grid_show" content="0"/>
  <meta name="grid_size" content="0.250000 0.250000"/>
  <meta name="grid_snap" content="0"/>
  <meta name="guide_color" content="0.435294 0.435294 1.000000"/>
  <meta name="guide_show" content="1"/>
  <meta name="guide_snap" content="0"/>
  <meta name="jack_offset" content="0.000000"/>
  <meta name="onion_skin" content="0"/>
  <meta name="onion_skin_future" content="0"/>
  <meta name="onion_skin_keyframes" content="1"/>
  <meta name="onion_skin_past" content="1"/>
  <keyframe time="0f" active="true">start</keyframe>
</canvas>"""



from test import *
import sangfroid
from sangfroid import T

def test_keyframes_simple():
    sif = get_animation('circles.sif')

    assert len(sif.keyframes)==6

    for k, expected in zip(sif.keyframes, EXPECTED_CIRCLES_KEYFRAMES):
        assert (k.time.frames, k.name, k.active) == expected

def test_keyframes_str():
    sif = get_animation('circles.sif')

    assert str(sif.keyframes)==EXPECTED_CIRCLES_STR
    assert repr(sif.keyframes)==EXPECTED_CIRCLES_STR

EXPECTED_CIRCLES_KEYFRAMES = [
        (0, "it's white", True),
        (20, "now it's pink", True),
        (25, "now it's mauve", True),
        (33, 'this one is inactive', False),
        (35, 'blue', True),
        (41, '', True),
        ]

EXPECTED_CIRCLES_STR = """Keyframes of [📂animation 'I like circles. They are round.']:
  - 0f "it's white"
  - 20f "now it's pink"
  - 1s 1f "now it's mauve"
  - 1s 9f 'this one is inactive' (inactive)
  - 1s 11f 'blue'
  - 1s 17f ''"""



"""
handle Synfig animations

Sangfroid allows you to load, change, and save Synfig animations
using Python.
"""
import sangfroid.layer
import sangfroid.value

from sangfroid.keyframe import *
from sangfroid.animation import *
from sangfroid.t import *

def open(filename):
    """
    Loads the Synfig file with the given filename.

    Args:
        filename (str): the name of the source file. Can be .sfg, .sif,
            or .sifz.

    Returns:
        Animation
    """
    result = Animation(filename)
    return result

__version__ = '0.1'

__all__ = [
        'layer',
        'Canvas',
        'Keyframe',
        'Animation',
        'T',

        'open',
        ]



import math
import functools
import re
import bs4

@functools.total_ordering
class T:
    """
    An instant during an animation.

    This can be expressed either as a number of frames or as a number
    of seconds, both measured from the start of the animation.
    In both cases, it's a float, usually but not always integral.

    The time may be specified in one of four ways, where `F` represents
    any decimal integer or float:

        - an `int` or `float`, giving the number of frames counting forwards
          from the start of the animation. But if the number is negative,
          it counts backwards from the end, in the usual Python indexing style.

        - `"Ff"`

        - `"Fs"`

        - `"Fs Ff"`, in which case the values are added together.

    For the ref parameter, you may supply the root Beautiful Soup
    tag of an animation, or any of that tag's descendants. If this parameter
    is unspecified, or `None`, but we need to calculate something from it,
    we throw `ValueError`.

    There are two reasons you might need to specify a ref:

        - If the specification is `int` or `float`, and it's negative;
          this is because we will need to look up the animation length.

        - If the specification is given as `"Fs"` or `"Fs Ff"`;
          this is because we need to look up the speed in frames per second.
          It's almost always 24. The reason we don't default to 24 is that
          it risks introducing subtle bugs which are only discovered when
          the actual frame rate isn't 24.

    However, if the time is zero (that is, at the start of the animation),
    you may write it as `"0s"` or `"0s 0f"` even if you don't supply
    the ref.  This doesn't allow you to write `"0s 2f"` without
    supplying the ref, even though the FPS wouldn't make
    a difference, because edge cases give rise to bugs.

    Ts compare numerically to other Ts, but unless the frame
    count is zero, they must have the same FPS.

    The number of seconds in a `"Fs"` or `"Fs Ff"` may be negative,
    which allows you to set times before the start of the animation.
    However, the number of *frames* in a `"Fs Ff"` specification
    may not be negative.

    In time specifications giving both seconds and frames, the number
    of frames may be equal to or higher than the FPS. (So, for example,
    `"2s 48f"` is a valid time even if `fps=24`, when it's equivalent
    to `"4s`"`.)

    This type is immutable and hashable.

    Attrs:
        _frames (float): distance in frames from the start of the animation
        _fps (float or None): speed in frames per second, if known
    """
    def __init__(
            self,
            value=0.0,
            ref = None,
            ):
 
        if ref is None:
            root_tag = None
        elif isinstance(ref, bs4.Tag):
            root_tag = _canvas_root(ref)
        elif hasattr(ref, 'tag'):
            root_tag = _canvas_root(getattr(ref, 'tag'))
        else:
            raise TypeError(
                    "'ref' must be bs4.Tag, a Layer subclass, or None, "
                    f"but not {type(ref)}.")

        if root_tag is None:
            self._fps = None
        else:
            self._fps = float(root_tag['fps'])
        
            if self._fps<=0.0:
                raise ValueError(f"FPS must be positive: {self._fps}")

        try:
            if isinstance(value, str):
                self._frames = self._parse_time_spec(value)
            else:
                self._frames = float(value)

                if self._frames < 0:
                    if root_tag is None:
                        raise ValueError("Negative frame counts require "
                                         "an anchored ref.")

                    animation_duration = 1+(
                            self._parse_time_spec(
                                root_tag.attrs['end-time'],
                                )
                            - 
                            self._parse_time_spec(
                                root_tag.attrs['begin-time'],
                                )
                            )

                    self._frames = animation_duration + self._frames

        except TypeError:
            raise ValueError(f"Invalid time specification: {value}")

        assert isinstance(self._frames, float)

    def _parse_time_spec(self, s):
        assert isinstance(s, str)

        def complain(problem):
            raise ValueError(
                    f"bad time specification: {s}\n{problem}")

        s = s.strip()

        seconds = None
        frames = 0.0

        found = TIMESPEC_RE.split(s)
        if found is None:
            complain("This doesn't look like a time specification at all.")

        parts = {}
        for spec in found:
            if spec.strip()=='':
                continue
            elif spec[-1] in parts:
                complain(
                        f"The '{spec[-1]}' part was specified "
                        "more than once.")

            try:
                parts[spec[-1]] = float(spec[:-1])
            except ValueError:
                complain(
                        f"'{spec[:-1]}' is not a valid number.")

        seconds = None

        for unit, size in [
                ('h', 60*60),
                ('m', 60),
                ('s', 1),
                ]:
            if unit in parts:
                seconds = (seconds or 0.0) + parts[unit]*size

        result = 0.0

        if seconds is not None and seconds!=0.0:

            if self._fps is None:
                complain("Time specifications in seconds require "
                         "an anchored ref.")
            result += seconds * self._fps

        if 'f' in parts:
            if seconds and parts['f']<0:
                complain("If you give a number of seconds, the number "
                         "of frames can't be negative.")

            seconds = seconds or 0

            if seconds<0:
                result -= parts['f']
            else:
                result += parts['f']

        return result

    @property
    def frames(self):
        """
        Time in frames.
        """
        return self._frames

    @property
    def fps(self):
        """
        Speed of the film, in frames per second. Always a positive float,
        or None if we don't know the speed.
        """
        return self._fps

    def __int__(self):
        return int(self._frames)

    def __float__(self):
        return self._frames

    @property
    def seconds(self):
        """
        Time in seconds.

        Raises:
            ValueError: if we don't know the FPS.
        """
        if self.fps is None:
            if self._frames==0.0:
                return 0.0
            else:
                raise ValueError(
                        "If you want to measure time in seconds, "
                        "you will need to specify the FPS.")
        return self._frames / self.fps

    def _compare(self, other, operator):

        if isinstance(other, self.__class__):
            if (
                    other.fps is not None and
                    self.fps is not None and
                    other.fps!=self.fps):
                raise ValueError(
                        "Comparison between two Ts with different FPS: "
                        f"{self}, {other}"
                        )

            other_f = other.frames
        elif isinstance(other, (float, int)):
            other_f = other
        else:
            return False

        return operator(self._frames, other_f)

    def __lt__(self, other):
        return self._compare(other, lambda a,b:a<b)

    def __eq__(self, other):
        return self._compare(other, lambda a,b:a==b)

    def __str__(self):
        if self._fps is None:
            return '%gf' % (self._frames, )

        frames = self._frames
        seconds = abs(self._frames/self._fps)

        result = []

        for unit, size in [
                ('h', 60*60),
                ('m', 60),
                ('s', 1),
                ]:

            if seconds>=size:
                result.append('%d%c' % (seconds//size, unit))
                seconds = seconds % size

        if seconds!=0 or result==[]:
            frames = abs(frames) % self._fps
            result.append('%gf' % (frames,))

        if self._frames<0:
            minus = '-'
        else:
            minus = ''

        result = minus + ' '.join(result)

        return result

    __repr__ = __str__

    def __hash__(self):
        if self._frames == 0:
            return 0

        s = (self._frames, self._fps)
        return hash(s)

# It doesn't matter that you can produce invalid decimals with this regex:
# that will be discovered when we do the float conversion.
TIMESPEC_RE = re.compile(
        r'(-?[0-9.]+[hmsf])'
        )

def _canvas_root(tag):
    parents = tag.find_parents()

    if len(parents)==0:
        return None
    elif len(parents)==1:
        root = tag
    else:
        root = parents[-2] # -1 is "document", the anon root tag

    if root.name=='canvas':
        return root
    else:
        return None

__all__ = [
        'T',
        ]



import bs4
from sangfroid.value.value import Value

RGBA = 'rgba'

@Value.handles_type()
class Color(Value):

    @property
    def value(self):
        result = '#' + (''.join([
                '%02x' % (
                    int(round(float(self._tag(dimension)[0].string)*255.0)),)
            for dimension in RGBA
            ]))

        if result.endswith('ff'):
            result = result[:-2]
        return result

    def as_tuple(self):
        result = tuple([
            float(self._tag(dimension)[0].string)
            for dimension in RGBA
            ])
        return result

    def __eq__(self, other):

        if isinstance(other, Color):
            return self.value == other.value
        elif isinstance(other, str):
            other = other.lower()
            if not other.startswith('#'):
                other = f'#{other}'
            if len(other)==9 and other.endswith('ff'):
                other = other[:-2]
            return str(self) == other
        elif isinstance(other, tuple):
            if len(other)==3:
                other = other+(1.0,)
            return self.as_tuple() == other
        else:
            return False

    @value.setter
    def value(self, v):

        def _raise_format_error():
            raise ValueError(
                    "Express a colour value as (r,g,b), or (r,g,b,a), or "
                    f"as a string of six or eight hex digits: {v}"
                    )

        if isinstance(v, tuple):

            if len(v)==0:
                result = (1.0, 1.0, 1.0, 1.0)
            elif len(v)==3:
                result = v + (1.0,)
            elif len(v)==4:
                result = v
            else:
                _raise_format_error()

            result = tuple([float(d) for d in result])

        elif isinstance(v, str):

            if v.startswith('#'):
                v = v[1:]

            if len(v)==6:
                v += 'FF'
            elif len(v)==8:
                pass
            else:
                _raise_format_error()

            result = tuple(
                round(int(v[i*2:i*2+2], 16)/255, 6)
                for i in range(4))

        elif isinstance(v, Color):

            result = v.as_tuple()

        else:
            _raise_format_error()

        assert isinstance(result, tuple), result
        assert [type(n) for n in result]==[float]*4, result

        self._tag.name = __class__.__name__.lower()
        self._tag.attrs = {}
        self._tag.clear()

        for i, dimension in enumerate(RGBA):
            dimension_tag = bs4.element.Tag(name=dimension)
            dimension_tag.string = '%.06f' % (result[i],)
            self._tag.append(dimension_tag)

    def as_python_expression(self):
        return self.as_tuple()



import bs4
from sangfroid.value.value import Value

@Value.handles_type()
class Vector(Value):

    # it may happen that other types occur, though I don't
    # know of any at present, or how we could know if they
    # applied to us
    our_type = float

    TAG_NAME = 'vector'
    FIELDS = ['x', 'y']

    @property
    def value(self):
        return dict(
                [(field.name,
                  field.string)
                 for field in self._tag.children
                 if isinstance(field, bs4.element.Tag)
                 ])

    def _raise_type_error(self):
        raise TypeError(
                "Vectors may be constructed as "
                f"{self.__class__.__name__}"
                f"({','.join(self.FIELDS)}), or "
                f"{self.__class__.__name__}"
                "(dict_of_members).")

    def __getattr__(self, f):
        if f.startswith('_'):
            raise AttributeError(f)

        s = self._tag.find(name=f)
        if s is None:
            raise AttributeError(f)
        v = self.our_type(s.text)
        return v

    @value.setter
    def value(self, v):

        if v is None:
            members = {}
        elif isinstance(v, tuple):
            if len(v)==0:
                members = {}
            elif len(v)==2:
                members = dict(zip(self.FIELDS, v))

                for m in members.values():
                    if not isinstance(m, (float, int)):
                        self._raise_type_error()

            else:
                self._raise_type_error()
        elif hasattr(v, 'items'):
            members = v
        else:
            self._raise_type_error()

        self._tag.name = self.TAG_NAME
        self._tag.attrs = {}

        for k, v in members.items():
            addendum = bs4.element.Tag(name=k)
            addendum.string = str(v)
            self._tag.append(addendum)

    def __getitem__(self, key):
        result = self.get(key, default=None)

        if result is None:
            raise KeyError(key)

        return result

    def get(self, key, default=None):
        if isinstance(key, int):
            key = self.keys()[key]

        v = [field.string
             for field in self._tag.children
             if isinstance(field, bs4.element.Tag)
             and field.name==key
             ]

        if len(v)==0:
            return default

        return self.our_type(v[0])

    # FIXME: All these methods are written in terms of self.value,
    # which is inefficient because all the values must be created
    # every time. They should be fixed to read self._tag themselves.

    def keys(self):
        return sorted(self.value.keys())

    def values(self):
        if self.is_animated:
            return None
        return [self.our_type(v) for v in self.value.values()]

    def items(self):
        return [(k, self.our_type(v)) for k,v in self.value.items()]

    def __len__(self):
        return len(self.value)

    def as_tuple(self):
        return tuple(
                [self.our_type(self.value[k])
                 for k in sorted(self.value.keys())]
                )

    def _str_inner(self):
        if sorted(self.value.keys())==['x', 'y']:
            return str(self.as_tuple())
        else:
            return str(self.value)

    def __eq__(self, other):
        try:
            if len(other)!=len(self):
                return False
        except TypeError:
            return False

        return all([left==right for left,right in zip(self, other)])

    def __iter__(self):
        for v in self.values():
            yield v

    def _get_empty_tag(cls):
        return super()._get_empty_tag(cls.TAG_NAME)

    @classmethod
    def _construct_from(cls, tag):

        def first_child_tag(t):
            tags = [n for n in t.children if isinstance(n, bs4.Tag)]
            if len(tags)==0:
                return None
            else:
                return tags[0]

        c = None
        start_tag = tag

        if tag.name==cls.ANIMATED:
            waypoint_tag = first_child_tag(tag)
            if waypoint_tag is None:
                # no way of telling; make the best guess
                c = X_Y
            else:
                start_tag = first_child_tag(waypoint_tag)

        if c is None:
            fields = set([
                f.name for f in start_tag.children
                if isinstance(f, bs4.Tag)
                ])

            if fields=={'x', 'y'}:
                c = X_Y
            else:
                raise TypeError(
                        "This tag isn't a vector type I know:\n"
                        f"  the fields are: {repr(fields)}.\n"
                        f"{tag}"
                        )

        assert c is not None
        return c(tag)

class X_Y(Vector):
    pass



import bs4
from sangfroid.value.value import Value
from sangfroid.value.simple import Angle
from sangfroid.value.vector import X_Y
from copy import copy

@Value.handles_type()
class Composite(Value):

    REQUIRED_KEYS = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.REQUIRED_KEYS is None:
            raise NotImplementedError()

    @classmethod
    def _get_empty_tag(cls):
        result = bs4.element.Tag(name='composite')
        result['type'] = cls.__name__.lower()
        return result

    @property
    def value(self):
        def name_and_value(n):

            name = n.name

            assert name!='layer'

            value_tags = [v
                 for v in n.children
                 if isinstance(v, bs4.element.Tag)
                 ]

            if len(value_tags)!=1:
                raise ValueError("Fields in composite types should only "
                                 f"have a single value: {n}")

            value = Value.from_tag(value_tags[0])

            return name, value

        result = dict(
                [name_and_value(field)
                 for field in self._tag.children
                 if isinstance(field, bs4.element.Tag)
                 ])

        if self.REQUIRED_KEYS is not None:
            assert result.keys()==self.REQUIRED_KEYS.keys()

        return result

    @value.setter
    def value(self, new_value):
        if not hasattr(new_value, 'items'):
            raise TypeError("The new value must be a dict.")

        for k,v in new_value.items():
            if self.REQUIRED_KEYS is not None:
                if k not in self.REQUIRED_KEYS:
                    raise KeyError(
                    f"{k} is not one of the keys we can accept. "
                    f"We can accept: "
                    f"{' '.join(sorted(self.REQUIRED_KEYS.keys()))}")

        seen = set()
        for field in self._tag.children:
            if not isinstance(field, bs4.element.Tag):
                continue

            if not field.name in new_value:
                continue

            # XXX TO HERE
            raise ValueError(new_value)

            seen.add(field.name)

        for remaining, v in new_value.items():
            if remaining in seen:
                continue

            new_subtag = bs4.Tag(name=remaining)
            new_subtag.append(copy(v.tag))
            new_subtag.append("\n")
            self.tag.append(new_subtag)

    def __getattr__(self, key):
        result = self.get(key, default=None)

        if result is None:
            raise AttributeError(key)

        return result

    def __getitem__(self, key):
        result = self.get(key=key, default=None)
        if result is None:
            raise KeyError(key)
        return result

    def get(self, key, default=None):
        found = [v
                 for v in self._tag.children
                 if isinstance(v, bs4.element.Tag)
                 and v.name==key
                 ]

        if len(found)==0:
            return default
        elif len(found)>1:
            raise ValueError(f"multiple values for {key}!")

        values = [v
                 for v in found[0].children
                 if isinstance(v, bs4.element.Tag)
                 ]

        if len(values)==0:
            return default
        elif len(values)>1:
            raise ValueError(f"multiple values for {key}!")
        elif values[0]==None:
            return default
        else:
            return Value.from_tag(values[0])

    def keys(self):
        # No point constructing all the values
        return [v.name
                 for v in self._tag.children
                 if isinstance(v, bs4.element.Tag)
                 ]

    def values(self):
        return self.value.values()

    def items(self):
        return self.value.items()

    def __len__(self):
        return len([v
                 for v in self._tag.children
                 if isinstance(v, bs4.element.Tag)
                 ])

    def __eq__(self, other):

        if not hasattr(other, 'items'):
            return False

        if len(other)!=len(self):
            return False

        for key, value in self.items():
            if other[key]!=value:
                return False

        return True

    def as_python_expression(self):
        value = self.value

        result = '{\n'
        for f,v in value.items():
            result += '     %40s: %s,\n' % (
                    repr(f),
                    v.as_python_expression(),
                    )
        result += (' '*36) + '}'

        return result

    @classmethod
    def _subfields(cls):
        return set(cls.REQUIRED_KEYS or ())

    @classmethod
    def _construct_from(cls, tag):
        try:
            subtype = tag['type']
        except KeyError:
            raise KeyError(
                    'This <composite> layer has no type parameter. '
                    'You probably wanted type="transformation".'
                    )

        if subtype.lower()=='transformation':
            return Transformation(tag)
        else:
            raise KeyError(
                    f'This <composite> layer has type="{subtype}", '
                    'which isn\'t a type I know.\n'
                    'You probably wanted type="transformation".'
                    )

class Transformation(Composite):
    REQUIRED_KEYS = {
            'offset': X_Y,
            'angle': Angle,
            'skew_angle': Angle,
            'scale': X_Y,
            }



from sangfroid.layer.layer import Layer
import sangfroid.value as v
import sangfroid.field as f

"""
These are the layer types that we haven't got to yet.
"""

@Layer.handles_type()
class Xor_Pattern(Layer):
    SYMBOL = 'X'
    ### {{{
    SYNFIG_VERSION = "0.1"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                        )
    amount               = f.ParamTagField(v.Real, 1.0,
                        )
    blend_method         = f.ParamTagField(v.BlendMethod, v.BlendMethod.COMPOSITE,
                        )
    origin               = f.ParamTagField(v.X_Y, (0.125, 0.125),
                        )
    size                 = f.ParamTagField(v.X_Y, (0.25, 0.25),
                        )

    ### }}}

@Layer.handles_type()
class Switch(Layer):
    SYMBOL = 'X'
    ### {{{
    SYNFIG_VERSION = "0.0"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                        )
    amount               = f.ParamTagField(v.Real, 1.0,
                        )
    blend_method         = f.ParamTagField(v.BlendMethod, v.BlendMethod.COMPOSITE,
                        )
    origin               = f.ParamTagField(v.X_Y, (0.0, 0.0),
                        )
    transformation       = f.ParamTagField(v.Transformation, {
                                         'offset': (0.0, 0.0),
                                          'angle': 0.0,
                                     'skew_angle': 0.0,
                                          'scale': (1.0, 1.0),
                                        },
                        )
    canvas               = f.SwitchCanvasField()
    tl                   = f.ParamTagField(v.X_Y, (-0.5333333611, 0.5333333611),
                        )
    br                   = f.ParamTagField(v.X_Y, (0.5333333611, -0.5333333611),
                        )
    c                    = f.ParamTagField(v.Integer, 1,
                        )
    gamma_adjust         = f.ParamTagField(v.Real, 1.0,
                        )
    filename             = f.ParamTagField(v.String, 'drop.png',
                        )
    time_offset          = f.ParamTagField(v.Time, 0,
                        )
    time_dilation        = f.ParamTagField(v.Real, 1.0,
                        )
    children_lock        = f.ParamTagField(v.Bool, True,
                        )
    outline_grow         = f.ParamTagField(v.Real, 0.0,
                        )
    layer_name           = f.ParamTagField(v.String, 'drop.png',
                        )
    layer_depth          = f.ParamTagField(v.Integer, -1,
                        )

    ### }}}

@Layer.handles_type()
class Super_Sample(Layer):
    SYMBOL = 'X'
    ### {{{
    SYNFIG_VERSION = "0.2"

    width                = f.ParamTagField(v.Integer, 2,
                        )
    height               = f.ParamTagField(v.Integer, 2,
                        )

    ### }}}

@Layer.handles_type()
class Sound(Layer):
    SYMBOL = '🔊'
    ### {{{
    SYNFIG_VERSION = "0.1"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                        )
    filename             = f.ParamTagField(v.String, 'laser.wav',
                        )
    delay                = f.ParamTagField(v.Time, 0,
                        )
    volume               = f.ParamTagField(v.Real, 1.0,
                        )

    ### }}}

@Layer.handles_type()
class Skeleton(Layer):
    SYMBOL = '💀'
    ### {{{
    SYNFIG_VERSION = "0.1"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                        )
    amount               = f.ParamTagField(v.Real, 0.5,
                        )
    name                 = f.ParamTagField(v.String, 'skeleton',
                        )
    bones                = f.NotImplementedField("Static_List",
                        )

    ### }}}

@Layer.handles_type()
class Plant(Layer):
    SYMBOL = '🪴'
    ### {{{
    SYNFIG_VERSION = "0.2"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                        )
    amount               = f.ParamTagField(v.Real, 1.0,
                        )
    blend_method         = f.ParamTagField(v.BlendMethod, v.BlendMethod.COMPOSITE,
                        )
    bline                = f.NotImplementedField("Bline",
                        )
    origin               = f.ParamTagField(v.X_Y, (0.0, 0.0),
                        )
    gradient             = f.ParamTagField(v.Gradient, {0.0:(1.0, 1.0, 1.0, 1.0),1.0:(0.0, 0.0, 0.0, 1.0)},
                        )
    split_angle          = f.ParamTagField(v.Angle, 10.0,
                        )
    gravity              = f.ParamTagField(v.X_Y, (0.0, -0.1000000015),
                        )
    velocity             = f.ParamTagField(v.Real, 0.3,
                        )
    perp_velocity        = f.ParamTagField(v.Real, 0.0,
                        )
    size                 = f.ParamTagField(v.Real, 0.015,
                        )
    size_as_alpha        = f.ParamTagField(v.Bool, False,
                        )
    reverse              = f.ParamTagField(v.Bool, True,
                        )
    step                 = f.ParamTagField(v.Real, 0.01,
                        )
    seed                 = f.ParamTagField(v.Integer, 1700432811,
                        )
    splits               = f.ParamTagField(v.Integer, 5,
                        )
    sprouts              = f.ParamTagField(v.Integer, 10,
                        )
    random_factor        = f.ParamTagField(v.Real, 0.2,
                        )
    drag                 = f.ParamTagField(v.Real, 0.1,
                        )
    use_width            = f.ParamTagField(v.Bool, True,
                        )

    ### }}}

@Layer.handles_type()
class Filter_Group(Layer):
    SYMBOL = 'X'
    ### {{{
    SYNFIG_VERSION = "0.1"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                        )
    amount               = f.ParamTagField(v.Real, 1.0,
                        )
    blend_method         = f.ParamTagField(v.BlendMethod, v.BlendMethod.STRAIGHT,
                        )
    origin               = f.ParamTagField(v.X_Y, (0.0, 0.0),
                        )
    transformation       = f.ParamTagField(v.Transformation, {
                                         'offset': (0.0, 0.0),
                                          'angle': 0.0,
                                     'skew_angle': 0.0,
                                          'scale': (1.0, 1.0),
                                        },
                        )
    canvas               = f.ParamTagField(v.Canvas, [],
                        )
    time_dilation        = f.ParamTagField(v.Real, 1.0,
                        )
    time_offset          = f.ParamTagField(v.Time, 0,
                        )
    children_lock        = f.ParamTagField(v.Bool, False,
                        )
    outline_grow         = f.ParamTagField(v.Real, 0.0,
                        )

    ### }}}

@Layer.handles_type()
class Duplicate(Layer):
    SYMBOL = 'X'
    ### {{{
    SYNFIG_VERSION = "0.1"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                        )
    amount               = f.ParamTagField(v.Real, 1.0,
                        )
    blend_method         = f.ParamTagField(v.BlendMethod, v.BlendMethod.COMPOSITE,
                        )
    index                = f.DuplicatesIndexField(None)

    ### }}}

@Layer.handles_type()
class Spiral_Gradient(Layer):
    SYMBOL = 'X'
    ### {{{
    SYNFIG_VERSION = "0.2"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                        )
    amount               = f.ParamTagField(v.Real, 1.0,
                        )
    blend_method         = f.ParamTagField(v.BlendMethod, v.BlendMethod.COMPOSITE,
                        )
    gradient             = f.ParamTagField(v.Gradient, {0.0:(1.0, 1.0, 1.0, 1.0),1.0:(0.0, 0.0, 0.0, 1.0)},
                        )
    center               = f.ParamTagField(v.X_Y, (0.0, 0.0),
                        )
    radius               = f.ParamTagField(v.Real, 0.5,
                        )
    angle                = f.ParamTagField(v.Angle, 0.0,
                        )
    clockwise            = f.ParamTagField(v.Bool, False,
                        )

    ### }}}

@Layer.handles_type()
class Radial_Gradient(Layer):
    SYMBOL = 'X'
    ### {{{
    SYNFIG_VERSION = "0.2"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                        )
    amount               = f.ParamTagField(v.Real, 1.0,
                        )
    blend_method         = f.ParamTagField(v.BlendMethod, v.BlendMethod.COMPOSITE,
                        )
    gradient             = f.ParamTagField(v.Gradient, {0.0:(1.0, 1.0, 1.0, 1.0),1.0:(0.0, 0.0, 0.0, 1.0)},
                        )
    center               = f.ParamTagField(v.X_Y, (0.0, 0.0),
                        )
    radius               = f.ParamTagField(v.Real, 0.5,
                        )
    loop                 = f.ParamTagField(v.Bool, False,
                        )
    zigzag               = f.ParamTagField(v.Bool, False,
                        )

    ### }}}

@Layer.handles_type()
class Noise(Layer):
    SYMBOL = 'X'
    ### {{{
    SYNFIG_VERSION = "0.0"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                        )
    amount               = f.ParamTagField(v.Real, 1.0,
                        )
    blend_method         = f.ParamTagField(v.BlendMethod, v.BlendMethod.COMPOSITE,
                        )
    gradient             = f.ParamTagField(v.Gradient, {0.0:(1.0, 1.0, 1.0, 1.0),1.0:(0.0, 0.0, 0.0, 1.0)},
                        )
    seed                 = f.ParamTagField(v.Integer, 1700432620,
                        )
    size                 = f.ParamTagField(v.X_Y, (1.0, 1.0),
                        )
    smooth               = f.ParamTagField(v.Integer, 2,
                        )
    detail               = f.ParamTagField(v.Integer, 4,
                        )
    speed                = f.ParamTagField(v.Real, 0.0,
                        )
    turbulent            = f.ParamTagField(v.Bool, False,
                        )
    do_alpha             = f.ParamTagField(v.Bool, False,
                        )
    super_sample         = f.ParamTagField(v.Bool, False,
                        )

    ### }}}

@Layer.handles_type()
class Linear_Gradient(Layer):
    SYMBOL = 'X'
    ### {{{
    SYNFIG_VERSION = "0.1"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                        )
    amount               = f.ParamTagField(v.Real, 1.0,
                        )
    blend_method         = f.ParamTagField(v.BlendMethod, v.BlendMethod.COMPOSITE,
                        )
    p1                   = f.ParamTagField(v.X_Y, (1.0, 1.0),
                        )
    p2                   = f.ParamTagField(v.X_Y, (-1.0, -1.0),
                        )
    gradient             = f.ParamTagField(v.Gradient, {0.0:(1.0, 1.0, 1.0, 1.0),1.0:(0.0, 0.0, 0.0, 1.0)},
                        )
    loop                 = f.ParamTagField(v.Bool, False,
                        )
    zigzag               = f.ParamTagField(v.Bool, False,
                        )

    ### }}}

@Layer.handles_type()
class Curve_Gradient(Layer):
    SYMBOL = 'X'
    ### {{{
    SYNFIG_VERSION = "0.0"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                        )
    amount               = f.ParamTagField(v.Real, 1.0,
                        )
    blend_method         = f.ParamTagField(v.BlendMethod, v.BlendMethod.COMPOSITE,
                        )
    origin               = f.ParamTagField(v.X_Y, (0.0, 0.0),
                        )
    width                = f.ParamTagField(v.Real, 0.0166666667,
                        )
    bline                = f.NotImplementedField("Bline",
                        )
    gradient             = f.ParamTagField(v.Gradient, {0.0:(1.0, 1.0, 1.0, 1.0),1.0:(0.0, 0.0, 0.0, 1.0)},
                        )
    loop                 = f.ParamTagField(v.Bool, False,
                        )
    zigzag               = f.ParamTagField(v.Bool, False,
                        )
    perpendicular        = f.ParamTagField(v.Bool, False,
                        )
    fast                 = f.ParamTagField(v.Bool, True,
                        )

    ### }}}

@Layer.handles_type()
class Conical_Gradient(Layer):
    SYMBOL = 'X'
    ### {{{
    SYNFIG_VERSION = "0.2"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                        )
    amount               = f.ParamTagField(v.Real, 1.0,
                        )
    blend_method         = f.ParamTagField(v.BlendMethod, v.BlendMethod.COMPOSITE,
                        )
    gradient             = f.ParamTagField(v.Gradient, {0.0:(1.0, 1.0, 1.0, 1.0),1.0:(0.0, 0.0, 0.0, 1.0)},
                        )
    center               = f.ParamTagField(v.X_Y, (0.0, 0.0),
                        )
    angle                = f.ParamTagField(v.Angle, 0.0,
                        )
    symmetric            = f.ParamTagField(v.Bool, False,
                        )

    ### }}}

@Layer.handles_type()
class Mandelbrot(Layer):
    SYMBOL = '🎨'
    ### {{{
    SYNFIG_VERSION = "0.2"

    iterations           = f.ParamTagField(v.Integer, 32,
                        )
    bailout              = f.ParamTagField(v.Real, 2.0,
                        )
    broken               = f.ParamTagField(v.Bool, False,
                        )
    distort_inside       = f.ParamTagField(v.Bool, True,
                        )
    shade_inside         = f.ParamTagField(v.Bool, True,
                        )
    solid_inside         = f.ParamTagField(v.Bool, False,
                        )
    invert_inside        = f.ParamTagField(v.Bool, False,
                        )
    gradient_inside      = f.ParamTagField(v.Gradient, {0.0:(1.0, 0.0, 0.0, 1.0),1.0:(1.0, 1.0, 0.0, 1.0)},
                        )
    gradient_offset_inside = f.ParamTagField(v.Real, 0.0,
                        )
    gradient_loop_inside = f.ParamTagField(v.Bool, True,
                        )
    distort_outside      = f.ParamTagField(v.Bool, True,
                        )
    shade_outside        = f.ParamTagField(v.Bool, True,
                        )
    solid_outside        = f.ParamTagField(v.Bool, False,
                        )
    invert_outside       = f.ParamTagField(v.Bool, False,
                        )
    gradient_outside     = f.ParamTagField(v.Gradient, {0.0:(0.0, 0.0, 0.0, 0.0),1.0:(0.0, 0.0, 0.0, 1.0)},
                        )
    smooth_outside       = f.ParamTagField(v.Bool, True,
                        )
    gradient_offset_outside = f.ParamTagField(v.Real, 0.0,
                        )
    gradient_scale_outside = f.ParamTagField(v.Real, 1.0,
                        )

    ### }}}

@Layer.handles_type()
class Julia(Layer):
    SYMBOL = '🎨'
    ### {{{
    SYNFIG_VERSION = "0.1"

    icolor               = f.ParamTagField(v.Color, (0.0, 0.0, 0.0, 1.0),
                        )
    ocolor               = f.ParamTagField(v.Color, (0.0, 0.0, 0.0, 1.0),
                        )
    color_shift          = f.ParamTagField(v.Angle, 0.0,
                        )
    iterations           = f.ParamTagField(v.Integer, 32,
                        )
    seed                 = f.ParamTagField(v.X_Y, (0.0, 0.0),
                        )
    bailout              = f.ParamTagField(v.Real, 2.0,
                        )
    distort_inside       = f.ParamTagField(v.Bool, True,
                        )
    shade_inside         = f.ParamTagField(v.Bool, True,
                        )
    solid_inside         = f.ParamTagField(v.Bool, False,
                        )
    invert_inside        = f.ParamTagField(v.Bool, False,
                        )
    color_inside         = f.ParamTagField(v.Bool, True,
                        )
    distort_outside      = f.ParamTagField(v.Bool, True,
                        )
    shade_outside        = f.ParamTagField(v.Bool, True,
                        )
    solid_outside        = f.ParamTagField(v.Bool, False,
                        )
    invert_outside       = f.ParamTagField(v.Bool, False,
                        )
    color_outside        = f.ParamTagField(v.Bool, False,
                        )
    color_cycle          = f.ParamTagField(v.Bool, False,
                        )
    smooth_outside       = f.ParamTagField(v.Bool, True,
                        )
    broken               = f.ParamTagField(v.Bool, False,
                        )

    ### }}}

@Layer.handles_type()
class Lumakey(Layer):
    SYMBOL = '🗝️'
    ### {{{
    SYNFIG_VERSION = "0.2"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                        )

    ### }}}

@Layer.handles_type()
class Halftone3(Layer):
    SYMBOL = '▓'
    ### {{{
    SYNFIG_VERSION = "0.0"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                        )
    amount               = f.ParamTagField(v.Real, 1.0,
                        )
    blend_method         = f.ParamTagField(v.BlendMethod, v.BlendMethod.STRAIGHT,
                        )
    size                 = f.ParamTagField(v.X_Y, (0.25, 0.25),
                        )
    type_                = f.ParamTagField(v.Integer, 0,
                        )
    subtractive          = f.ParamTagField(v.Bool, True,
                        )
    color                = f.ParamArrayField(v.Color, (0.0, 1.0, 1.0, 1.0),
                        )
    tone                 = f.ParamArrayField(v.Tone, default=None)

    ### }}}

@Layer.handles_type()
class Halftone2(Layer):
    SYMBOL = '▒'
    ### {{{
    SYNFIG_VERSION = "0.0"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                        )
    amount               = f.ParamTagField(v.Real, 1.0,
                        )
    blend_method         = f.ParamTagField(v.BlendMethod, v.BlendMethod.STRAIGHT,
                        )
    origin               = f.ParamTagField(v.X_Y, (0.0, 0.0),
                        )
    angle                = f.ParamTagField(v.Angle, 0.0,
                        )
    size                 = f.ParamTagField(v.X_Y, (0.25, 0.25),
                        )
    color_light          = f.ParamTagField(v.Color, (1.0, 1.0, 1.0, 1.0),
                        )
    color_dark           = f.ParamTagField(v.Color, (0.0, 0.0, 0.0, 1.0),
                        )
    type_                = f.ParamTagField(v.Integer, 0,
                        )

    ### }}}

@Layer.handles_type()
class Colorcorrect(Layer):
    SYMBOL = '👍'
    ### {{{
    SYNFIG_VERSION = "0.1"

    hue_adjust           = f.ParamTagField(v.Angle, 0.0,
                        )
    brightness           = f.ParamTagField(v.Real, 0.0,
                        )
    contrast             = f.ParamTagField(v.Real, 1.0,
                        )
    exposure             = f.ParamTagField(v.Real, 0.0,
                        )
    gamma                = f.ParamTagField(v.Real, 1.0,
                        )

    ### }}}

@Layer.handles_type()
class Clamp(Layer):
    SYMBOL = '🗜️'
    ### {{{
    SYNFIG_VERSION = "0.2"

    invert_negative      = f.ParamTagField(v.Bool, False,
                        )
    clamp_ceiling        = f.ParamTagField(v.Bool, True,
                        )
    ceiling              = f.ParamTagField(v.Real, 1.0,
                        )
    floor                = f.ParamTagField(v.Real, 0.0,
                        )

    ### }}}

@Layer.handles_type()
class Chromakey(Layer):
    SYMBOL = '🔑'
    ### {{{
    SYNFIG_VERSION = "0.1"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                        )
    key_color            = f.ParamTagField(v.Color, (0.0, 1.0, 0.0, 1.0),
                        )
    lower_bound          = f.ParamTagField(v.Real, 0.001,
                        )
    upper_bound          = f.ParamTagField(v.Real, 0.001,
                        )
    supersample_width    = f.ParamTagField(v.Integer, 2,
                        )
    supersample_height   = f.ParamTagField(v.Integer, 2,
                        )
    desaturate           = f.ParamTagField(v.Bool, True,
                        )
    invert               = f.ParamTagField(v.Bool, False,
                        )

    ### }}}

@Layer.handles_type()
class Simple_Circle(Layer):
    SYMBOL = 'X'
    ### {{{
    SYNFIG_VERSION = "0.1"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                        )
    amount               = f.ParamTagField(v.Real, 1.0,
                        )
    blend_method         = f.ParamTagField(v.BlendMethod, v.BlendMethod.COMPOSITE,
                        )
    color                = f.ParamTagField(v.Color, (1.0, 1.0, 1.0, 1.0),
                        )
    center               = f.ParamTagField(v.X_Y, (0.0, 0.0),
                        )
    radius               = f.ParamTagField(v.Real, 0.5,
                        )

    ### }}}

@Layer.handles_type()
class Metaballs(Layer):
    SYMBOL = 'X'
    ### {{{
    SYNFIG_VERSION = "0.1"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                        )
    amount               = f.ParamTagField(v.Real, 1.0,
                        )
    blend_method         = f.ParamTagField(v.BlendMethod, v.BlendMethod.COMPOSITE,
                        )
    gradient             = f.ParamTagField(v.Gradient, {0.0:(1.0, 1.0, 1.0, 1.0),1.0:(0.0, 0.0, 0.0, 1.0)},
                        )
    centers              = f.NotImplementedField("Dynamic_List",
                        )
    radii                = f.NotImplementedField("Dynamic_List",
                        )
    weights              = f.NotImplementedField("Dynamic_List",
                        )
    threshold            = f.ParamTagField(v.Real, 0.0,
                        )
    threshold2           = f.ParamTagField(v.Real, 1.0,
                        )
    positive             = f.ParamTagField(v.Bool, False,
                        )

    ### }}}

@Layer.handles_type()
class Warp(Layer):
    SYMBOL = 'X'
    ### {{{
    SYNFIG_VERSION = "0.2"

    src_tl               = f.ParamTagField(v.X_Y, (-2.0, 2.0),
                        )
    src_br               = f.ParamTagField(v.X_Y, (2.0, -2.0),
                        )
    dest_tl              = f.ParamTagField(v.X_Y, (-1.7999999523, 2.0999999046),
                        )
    dest_tr              = f.ParamTagField(v.X_Y, (1.7999999523, 2.0999999046),
                        )
    dest_br              = f.ParamTagField(v.X_Y, (2.2000000477, -2.0),
                        )
    dest_bl              = f.ParamTagField(v.X_Y, (-2.2000000477, -2.0),
                        )
    clip                 = f.ParamTagField(v.Bool, True,
                        )
    interpolation        = f.ParamTagField(v.Integer, 3,
                        )

    ### }}}

@Layer.handles_type()
class Twirl(Layer):
    SYMBOL = 'X'
    ### {{{
    SYNFIG_VERSION = "0.1"

    center               = f.ParamTagField(v.X_Y, (0.0, 0.0),
                        )
    radius               = f.ParamTagField(v.Real, 1.0,
                        )
    rotations            = f.ParamTagField(v.Angle, 0.0,
                        )
    distort_inside       = f.ParamTagField(v.Bool, True,
                        )
    distort_outside      = f.ParamTagField(v.Bool, False,
                        )

    ### }}}

@Layer.handles_type()
class Stretch(Layer):
    SYMBOL = 'X'
    ### {{{
    SYNFIG_VERSION = "0.1"

    amount               = f.ParamTagField(v.X_Y, (1.0, 1.0),
                        )
    center               = f.ParamTagField(v.X_Y, (0.0, 0.0),
                        )

    ### }}}

@Layer.handles_type()
class Spherize(Layer):
    SYMBOL = 'X'
    ### {{{
    SYNFIG_VERSION = "0.2"

    center               = f.ParamTagField(v.X_Y, (0.0, 0.0),
                        )
    radius               = f.ParamTagField(v.Real, 1.0,
                        )
    amount               = f.ParamTagField(v.Real, 1.0,
                        )
    clip                 = f.ParamTagField(v.Bool, False,
                        )
    type_                = f.ParamTagField(v.Integer, 0,
                        )

    ### }}}

@Layer.handles_type()
class Skeleton_Deformation(Layer):
    SYMBOL = 'X'
    ### {{{
    SYNFIG_VERSION = "0.2"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                        )
    amount               = f.ParamTagField(v.Real, 1.0,
                        )
    blend_method         = f.ParamTagField(v.BlendMethod, v.BlendMethod.STRAIGHT,
                        )
    bones                = f.NotImplementedField("Static_List",
                        )
    point1               = f.ParamTagField(v.X_Y, (-4.0, 4.0),
                        )
    point2               = f.ParamTagField(v.X_Y, (4.0, -4.0),
                        )
    x_subdivisions       = f.ParamTagField(v.Integer, 32,
                        )
    y_subdivisions       = f.ParamTagField(v.Integer, 32,
                        )

    ### }}}

@Layer.handles_type()
class Noise_Distort(Layer):
    SYMBOL = 'X'
    ### {{{
    SYNFIG_VERSION = "0.0"

    z_depth              = f.ParamTagField(v.Real, 0.0,
                        )
    amount               = f.ParamTagField(v.Real, 1.0,
                        )
    blend_method         = f.ParamTagField(v.BlendMethod, v.BlendMethod.STRAIGHT,
                        )
    displacement         = f.ParamTagField(v.X_Y, (0.25, 0.25),
                        )
    size                 = f.ParamTagField(v.X_Y, (1.0, 1.0),
                        )
    seed                 = f.ParamTagField(v.Integer, 1700432512,
                        )
    smooth               = f.ParamTagField(v.Integer, 2,
                        )
    detail               = f.ParamTagField(v.Integer, 4,
                        )
    speed                = f.ParamTagField(v.Real, 0.0,
                        )
    turbulent            = f.ParamTagField(v.Bool, False,
                        )

    ### }}}

@Layer.handles_type()
class Inside_Out(Layer):
    SYMBOL = 'X'
    ### {{{
    SYNFIG_VERSION = "0.1"

    origin               = f.ParamTagField(v.X_Y, (0.0, 0.0),
                        )

    ### }}}

@Layer.handles_type()
class Curve_Warp(Layer):
    SYMBOL = 'X'
    ### {{{
    SYNFIG_VERSION = "0.0"

    origin               = f.ParamTagField(v.X_Y, (0.0, 0.0),
                        )
    perp_width           = f.ParamTagField(v.Real, 1.0,
                        )
    start_point          = f.ParamTagField(v.X_Y, (-2.5, -0.5),
                        )
    end_point            = f.ParamTagField(v.X_Y, (2.5, -0.3000000119),
                        )
    bline                = f.NotImplementedField("Bline",
                        )
    fast                 = f.ParamTagField(v.Bool, True,
                        )

    ### }}}





import logging
from enum import Enum
import bs4
import sangfroid.value as v

logger = logging.getLogger('sangfroid')

class Field:
    """
    An attribute of a Layer, which represents a value within
    a Synfig layer class.

    For example, a circle has a radius, so Circle layers
    have a Field attribute named "radius".

    Field itself is an abstract superclass, because field
    data can be fetched in various ways. Each way has its
    own subclass, defined later in this file.

    Attributes:
        type_ (type): the permissible type of this value.
            Note the underscore, to avoid a clash with the
            reserved word. This value can live either within
            builtins (such as float or int), or within
            the `sangfroid.value` package..
        default (type_): the default value. This is what
            you get if you create a new Layer and
            don't specify any other value.
        name (str): the name of this layer, such as "radius".
            Usually we figure this out automatically, but
            you can specify it in the constructor because
            sometimes the name we want to use is a
            reserved word.
        owner (any): the class this Field lives in.
        doc (str): the docstring. (Should this live here?)

    """
    def __init__(self,
                 type_,
                 default,
                 name = None,
                 doc = None,
                 ):
        self.type_ = type_
        self.default = default
        self.name = name
        self.owner = None
        self.doc = doc

        self.__doc__ = doc or ''
        self.__doc__ += f"\n\nType: {type_.__name__}"
        if default is not None:
            # yes, this is the right way round; think about it
            self.__doc__ += ' or None'

        if issubclass(type_, Enum):
            self.__doc__ += '\n\nPossible values (integer or constants):\n\n'

            max_length = max([len(s) for s in type_.__members__],
                             default=0)

            self.__doc__ += '\n'.join([
                '%4d %*s %s' % (
                    i,
                    max_length,
                    s,
                    '...', # doc, FIXME
                    )
                for i, s in enumerate(type_.__members__)])

    def __set_name__(self, owner, name):
        self.owner = owner
        if self.name is None:
            self.name = name#.replace('_', '-')

    def __get__(self, obj, obj_type=None):
        raise NotImplementedError()

    def __set__(self, obj, value):
        raise NotImplementedError()

    def __str__(self):

        result = f'[{self.__class__.__name__}'

        result += '%20s of %20s (%20s)' % (
                self.name,
                self.owner.__name__,
                self.type_,)

        result += ']'

        return result

    __repr__ = __str__

class NotImplementedField(Field):
    """
    A Field we haven't implemented yet.

    You're welcome to load or save the layer, but attempting
    to access the value within Python will throw
    NotImplementedError.

    Attributes:
        typename (str): the name of the type which doesn't exist.
    """
    def __init__(self, typename = None):
        super().__init__(
                type_ = str,
                default = None,
                )
        self.typename = typename

    def _throw_not_implemented(self):
        if self.typename is None:
            raise NotImplementedError(
                    f"The type {self.owner} has not been "
                    "implemented yet. Patches welcome.")
        else:
            raise NotImplementedError(
                    f"The type {self.owner} requires a value "
                    f"of type f{self.type_}, but that hasn't been "
                    "implemented yet. Patches welcome.")

    def __get__(self, obj, obj_type=None):
        self._throw_not_implemented()

    def __set__(self, obj, value):
        self._throw_not_implemented()

class TagAttrField(Field):
    """
    A Field found in the attributes of a tag.

    For example, in
    ```
    <layer active="true" ...>
    ```

    the field `active` is a tag attribute field.

    Usually, the `type_` of a tag attribute field must be `str`,
    because it's encoded in an XML attribute. If you set the
    attribute `type_override` to another type, the string
    will be coerced to and from that type.

    Attributes:
        type_override (builtin type): the type you want to set and get,
            even though the attribute itself is a string.
            If you pass None to the constructor, this will
            be set to the same value as `type_`.
    """

    def __init__(self,
                 *args,
                 type_override = None,
                 **kwargs):

        super().__init__(*args, **kwargs)

        self.type_override = type_override or self.type_
        assert self.type_override.__module__=='builtins', self.type_override

    def __get__(self, obj, obj_type=None):

        if self.name not in obj._tag.attrs:
            logger.debug("no %s field; returning default: %s",
                         self.name,
                         self.default,
                         )
            return self.default

        value = obj._tag.get(self.name)
        logger.debug("%s field is %s",
                     self.name,
                     value,
                     )

        if value is None:
            return None
        elif issubclass(self.type_override, bool):
            return str(value).lower()=='true'
        else:
            return self.type_override(value)

    def __set__(self, obj, value):
        if issubclass(self.type_override, bool):
            if value:
                value = 'true'
            else:
                value = 'false'
        else:
            value = self.type_override(value)

        obj._tag[self.name] = value

class ParamTagField(Field):
    """
    A Field which lives in a `<param>` tag within its layer.

    The `type_` field cannot be a builtin type. It should be
    a class from the `sangfroid.value` package, because we'll
    need to encode and decode it from XML.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        assert self.type_.__module__!='builtins', self.type_

    def _get_value(self, obj):
        holder = obj._tag.find('param',
                               attrs={
                                   'name': self.name,
                                   },
                               )

        if holder is None:
            raise AttributeError(self.name)

        contents = [t for t in holder.children
                    if isinstance(t, bs4.Tag)]
        if len(contents)!=1:
            raise ValueError(
                    "Param tags should have one child tag, which holds "
                    "the value of the param. However, this tag:\n\n"
                    f"{obj.tag}\n\n"
                    f"contains {len(contents)}.")

        result = v.Value.from_tag(contents[0])
        return result

    def __get__(self, obj, obj_type=None):
        result = self._get_value(obj)
        assert isinstance(result, self.type_), f"{result} {type(result)} {self.type_}"
        return result

    def __set__(self, obj, value):
        value_obj = self._get_value(obj)
        value_obj.value = value

class TagField(Field):
    """
    A Field representing the XML tag of the layer itself.

    Read-only.
    """
    def __init__(self):
        super().__init__(
                name = 'tag',
                type_ = bs4.Tag,
                default = None,
                doc = """The BeautifulSoup tag behind this item.""",
                )
 
    def __get__(self, obj, obj_type=None):
        return obj._tag

    def __set__(self, obj, value):
        raise KeyError("You can't put a different tag into an object.")

class NamedChildField(Field):
    """
    A Field which lives in a child tag of the layer, where the
    name of the child tag is the same as the field's name.

    For example, if the name was "wombat", the XML might look
    like this:

    ```
    <layer>
      <wombat>
        whatever the value is
      </wombat>
    </layer>
    """
    def __init__(self, type_, default, name=None, doc=None):
        super().__init__(
                type_ = type_,
                default = default,
                name = name,
                doc = doc,
                )

    def get_subtag_for_obj(self, obj):
        return obj._tag.find(self.name)
 
    def __get__(self, obj, obj_type=None):

        subtag = self.get_subtag_for_obj(obj)

        if subtag is None:
            return ''
        else:
            return subtag.string

    def __set__(self, obj, value):

        subtag = self.get_subtag_for_obj(obj)

        subtag.string = value

"""
A field which accesses a key in the value of another field.

For example, `g.offset` might be short for
`g.transformation['offset']`.

Use the `include_from` decorator classmethod to introduce
these fields from an existing field.
"""
class ShortcutField(Field):
    def __init__(self, upstream, upstream_field):
        self.upstream = upstream
        self.upstream_field = upstream_field

    def __get__(self, obj, obj_type=None):
        v = self.upstream.__get__(obj, obj_type)
        result = v[self.upstream_field]
        return result

    def __set__(self, obj, value):
        self.upstream.__set__(obj,
                              {
                                  self.upstream_field: value,
                                  })

    """
    Decorator. Imports all the keys of the named field as
    ShortcutFields.

    Args:
        name (str): the name of the method to access

    Raises:
        AttributeError: if there is no field called `name`
        KeyError: if the keys already exist within the class
    """
    @classmethod
    def include_from(cls, name):
        def inner(c):
            try:
                a = c.__getattribute__(c, name)
            except AttributeError:
                raise AttributeError(
                        f"{c} has no attribute called {repr(name)}.\n"
                        f"But it does have: {' '.join(dir(c))}.")

            for subfield in a.type_._subfields():
                try:
                    print(c.__getattribute__(c, subfield))
                    raise KeyError(subfield)
                except AttributeError:
                    pass
                setattr(c, subfield,
                        cls(a, subfield))
            return c
        return inner

class BlendMethodField(ParamTagField):
    def __init__(self, foo):
        super().__init__(v.Real, -1)

class ParamArrayField(ParamTagField):
    pass

class SwitchCanvasField(NotImplementedField):
    pass

class DuplicatesIndexField(NotImplementedField):
    pass



from sangfroid.value.value import Value

class Tbd(Value):
    """
    All types which are not yet implemented.
    """
    @property
    def value(self):
        raise NotImplementedError(
                f"{self.__class__.__name__} is not yet implemented.")

@Value.handles_type()
class Dynamic_List(Tbd):
    pass

@Value.handles_type()
class Static_List(Tbd):
    pass

@Value.handles_type()
class Wplist(Tbd):
    pass

@Value.handles_type()
class Dilist(Tbd):
    pass

@Value.handles_type()
class Bline(Tbd):
    pass

@Value.handles_type()
class Tone(Tbd):
    pass



import bs4
from sangfroid.value.value import Value
from sangfroid.t import T

@Value.handles_type()
class String(Value):
    our_type = str

    @property
    def value(self):
        result = str(self._tag.string)
        return result

    @value.setter
    def value(self, v):
        self._tag.string = str(v)

    def as_python_expression(self):
        return repr(self.value)
